from collections.abc import Iterator
from typing import TYPE_CHECKING, Optional

from pydantic import StrictStr

from snowflake.core import PollingOperation
from snowflake.core._common import AccountObjectCollectionParent, CreateMode, DeleteMode, ObjectReferenceMixin
from snowflake.core._generated.api_client import StoredProcApiClient
from snowflake.core._internal.telemetry import api_telemetry
from snowflake.core._internal.utils import deprecated
from snowflake.core._operation import PollingOperations
from snowflake.core._utils import (
    tag_assignment_to_tag_tuple,
    tag_resource_to_tag_reference,
    tag_tuple_to_tag_assignment,
)
from snowflake.core.role._generated import TagAssignment, TagReference
from snowflake.core.role._generated.api import RoleApi
from snowflake.core.role._generated.models.containing_scope import ContainingScope
from snowflake.core.role._generated.models.grant import Grant
from snowflake.core.role._generated.models.grant_of import GrantOf
from snowflake.core.role._generated.models.grant_on import GrantOn
from snowflake.core.role._generated.models.role import RoleModel as Role
from snowflake.core.role._generated.models.securable import Securable
from snowflake.core.tag import TagResource, TagValue


if TYPE_CHECKING:
    from snowflake.core import Root


class RoleCollection(AccountObjectCollectionParent["RoleResource"]):
    """Represents the collection operations on the Snowflake Role resource.

    With this collection, you can create, iterate through, and search for roles that you have access to in the
    current context.

    Examples
    ________
    Creating a role instance:

    >>> role_collection = root.roles
    >>> role_collection.create(Role(name="test-role", comment="samplecomment"))
    """

    def __init__(self, root: "Root") -> None:
        super().__init__(root, ref_class=RoleResource)
        self._api = RoleApi(
            root=self.root, resource_class=self._ref_class, sproc_client=StoredProcApiClient(root=self.root)
        )

    @api_telemetry
    def create(self, role: Role, *, mode: CreateMode = CreateMode.error_if_exists) -> "RoleResource":
        """Create a role in Snowflake.

        Parameters
        __________
        role: Role
            The ``Role`` object, together with the ``Role``'s properties:
            name ; comment is optional
        mode: CreateMode, optional
            One of the following enum values.

            ``CreateMode.error_if_exists``: Throw an :class:`snowflake.core.exceptions.ConflictError`
            if the role already exists in Snowflake.  Equivalent to SQL ``create role <name> ...``.

            ``CreateMode.or_replace``: Replace if the role already exists in Snowflake. Equivalent to SQL
            ``create or replace role <name> ...``.

            ``CreateMode.if_not_exists``: Do nothing if the role already exists in Snowflake.
            Equivalent to SQL ``create role <name> if not exists...``

            Default is ``CreateMode.error_if_exists``.


        Examples
        ________
        Creating a role, replacing any existing role with the same name:

        >>> role = Role(name="test-role", comment="samplecomment")
        >>> role_ref = root.roles.create(role, mode=CreateMode.or_replace)
        """
        real_mode = CreateMode[mode].value
        self._api.create_role(role._to_model(), StrictStr(real_mode), async_req=False)
        return self[role.name]

    @api_telemetry
    def create_async(
        self, role: Role, *, mode: CreateMode = CreateMode.error_if_exists
    ) -> PollingOperation["RoleResource"]:
        """An asynchronous version of :func:`create`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """  # noqa: D401
        real_mode = CreateMode[mode].value
        future = self._api.create_role(role._to_model(), StrictStr(real_mode), async_req=True)
        return PollingOperation(future, lambda _: self[role.name])

    def iter(
        self,
        *,
        like: Optional[str] = None,
        limit: Optional[int] = None,
        starts_with: Optional[str] = None,
        from_name: Optional[str] = None,
    ) -> Iterator[Role]:
        """Iterate through ``Role`` objects from Snowflake, filtering on any optional 'like' pattern.

        Parameters
        _________
        like: str, optional
            A case-insensitive string functioning as a filter, with support for SQL
            wildcard characters (% and _).
        starts_with: str, optional
            String used to filter the command output based on the string of characters that appear
            at the beginning of the object name. Uses case-sensitive pattern matching.
        show_limit: int, optional
            Limit of the maximum number of rows returned by iter(). The default is ``None``, which behaves equivalently
            to show_limit=10000. This value must be between ``1`` and ``10000``.
        from_name: str, optional
            Fetch rows only following the first row whose object name matches
            the specified string. This is case-sensitive and does not have to be the full name.

        Examples
        ________
        Showing all roles that you have access to see:

        >>> roles = role_collection.iter()

        Showing information of the exact role you want to see:

        >>> roles = role_collection.iter(like="your-role-name")

        Showing roles starting with 'your-role-name-':

        >>> roles = role_collection.iter(like="your-role-name-%")
        >>> roles = role_collection.iter(starts_with="your-role-name")

        Using a for loop to retrieve information from iterator:

        >>> for role in roles:
        ...     print(role.name, role.comment)
        """
        roles = self._api.list_roles(
            StrictStr(like) if like else None,
            StrictStr(starts_with) if starts_with else None,
            limit,
            from_name=from_name,
            async_req=False,
        )

        return map(Role._from_model, iter(roles))

    def iter_async(
        self,
        *,
        like: Optional[str] = None,
        limit: Optional[int] = None,
        starts_with: Optional[str] = None,
        from_name: Optional[str] = None,
    ) -> PollingOperation[Iterator[Role]]:
        """An asynchronous version of :func:`iter`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """  # noqa: D401
        future = self._api.list_roles(
            StrictStr(like) if like else None,
            StrictStr(starts_with) if starts_with else None,
            limit,
            from_name=from_name,
            async_req=True,
        )
        return PollingOperation(future, lambda rest_models: map(Role._from_model, iter(rest_models)))


class RoleResource(ObjectReferenceMixin[RoleCollection]):
    """Represents a reference to a Snowflake role.

    With this role reference, you can delete a role, as well as perform certain actions on it.
    """

    def __init__(self, name: str, collection: RoleCollection) -> None:
        self.name = name
        self.collection = collection

    @api_telemetry
    @deprecated("drop")
    def delete(self) -> None:
        self.drop()

    @api_telemetry
    def drop(self, if_exists: Optional[bool] = None) -> None:
        """Drop this role.

        Parameters
        __________
        if_exists: bool, optional
            Check the existence of this role before dropping it.
            Default is ``None``, which is equivalent to ``False``.

        Examples
        ________
        Deleting a role using its reference:

        >>> role_reference.drop()
        """
        self.collection._api.delete_role(self.name, if_exists, async_req=False)

    @api_telemetry
    def drop_async(self, if_exists: Optional[bool] = None) -> PollingOperation[None]:
        """An asynchronous version of :func:`drop`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """  # noqa: D401
        future = self.collection._api.delete_role(self.name, if_exists, async_req=True)
        return PollingOperations.empty(future)

    @api_telemetry
    def grant_role(self, role_type: str, role: Securable) -> None:
        """
        Grant a role to this role.

        Parameters
        __________
        role_type: str
            The type of role which would be granted.
        role: Securable
            The role which would be granted.

        Examples
        ________
        Using a role reference to grant a role to it:

        >>> reference_role.grant("role", Securable(name="test_role"))
        """
        grant = Grant(securable_type=role_type, securable=role)
        self.collection._api.grant_privileges(self.name, grant, async_req=False)

    @api_telemetry
    def grant_role_async(self, role_type: str, role: Securable) -> PollingOperation[None]:
        """An asynchronous version of :func:`grant_role`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """  # noqa: D401
        grant = Grant(securable_type=role_type, securable=role)
        future = self.collection._api.grant_privileges(self.name, grant, async_req=True)
        return PollingOperations.empty(future)

    @api_telemetry
    def grant_privileges(
        self,
        privileges: list[str],
        securable_type: str,
        securable: Optional[Securable] = None,
        grant_option: Optional[bool] = None,
    ) -> None:
        """
        Grant privileges on a securable to this role.

        Parameters
        __________
        privileges: list[str]
            The list of privileges to be granted.
        securable_type: str
            The type of securable on which the privileges would be granted.
        securable: Securable, optional
            The securable on which the privileges would be granted. Default is None.
        grant_option: bool, optional
            If True, the grantee can grant the privilege to others. Default is None which means False.

        Examples
        ________
        Using a role reference to grant privileges to it:

        >>> reference_role.grant_privileges(["CREATE", "USAGE"], "database", Securable(database="test_db"))
        """
        grant = Grant(
            privileges=privileges, securable_type=securable_type, securable=securable, grant_option=grant_option
        )
        self.collection._api.grant_privileges(self.name, grant, async_req=False)

    @api_telemetry
    def grant_privileges_async(
        self,
        privileges: list[str],
        securable_type: str,
        securable: Optional[Securable] = None,
        grant_option: Optional[bool] = None,
    ) -> PollingOperation[None]:
        """An asynchronous version of :func:`grant_privileges`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """  # noqa: D401
        grant = Grant(
            privileges=privileges, securable_type=securable_type, securable=securable, grant_option=grant_option
        )
        future = self.collection._api.grant_privileges(self.name, grant, async_req=True)
        return PollingOperations.empty(future)

    @api_telemetry
    def grant_privileges_on_all(
        self,
        privileges: list[str],
        securable_type: str,
        containing_scope: ContainingScope,
        grant_option: Optional[bool] = None,
    ) -> None:
        """
        Grant privileges on all securables matching the criteria to this role.

        Parameters
        __________
        privileges: list[str]
            The list of privileges to be granted.
        securable_type: str
            The type of securable on which the privileges would be granted.
        containing_scope: ContainingScope
            The criteria to match the securables.
        grant_option: bool, optional
            If True, the grantee can grant the privilege to others. Default is None which means False.

        Examples
        ________
        Using a role reference to grant privileges on all schemas in a database to it:

        >>> reference_role.grant_privileges_on_all(
        ...     ["CREATE", "USAGE"], "schema", ContainingScope(database="test_db")
        ... )
        """
        grant = Grant(
            privileges=privileges,
            securable_type=securable_type,
            containing_scope=containing_scope,
            grant_option=grant_option,
        )
        self.collection._api.grant_privileges(self.name, grant, async_req=False)

    @api_telemetry
    def grant_privileges_on_all_async(
        self,
        privileges: list[str],
        securable_type: str,
        containing_scope: ContainingScope,
        grant_option: Optional[bool] = None,
    ) -> PollingOperation[None]:
        """An asynchronous version of :func:`grant_privileges_on_all`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """  # noqa: D401
        grant = Grant(
            privileges=privileges,
            securable_type=securable_type,
            containing_scope=containing_scope,
            grant_option=grant_option,
        )
        future = self.collection._api.grant_privileges(self.name, grant, async_req=True)
        return PollingOperations.empty(future)

    @api_telemetry
    def grant_future_privileges(
        self,
        privileges: list[str],
        securable_type: str,
        containing_scope: ContainingScope,
        grant_option: Optional[bool] = None,
    ) -> None:
        """
        Grant privileges on all future securables matching the criteria to this role.

        Parameters
        __________
        privileges: list[str]
            The list of privileges to be granted.
        securable_type: str
            The type of securable on which the privileges would be granted.
        containing_scope: ContainingScope
            The criteria to match the securables.
        grant_option: bool, optional
            If True, the grantee can grant the privilege to others. Default is None which means False.

        Examples
        ________
        Using a role reference to grant privileges on all future schemas in a database to it:

        >>> reference_role.grant_future_privileges(
        ...     ["CREATE", "USAGE"], "schema", ContainingScope(database="test_db")
        ... )
        """
        grant = Grant(
            privileges=privileges,
            securable_type=securable_type,
            containing_scope=containing_scope,
            grant_option=grant_option,
        )
        self.collection._api.grant_future_privileges(self.name, grant, async_req=False)

    @api_telemetry
    def grant_future_privileges_async(
        self,
        privileges: list[str],
        securable_type: str,
        containing_scope: ContainingScope,
        grant_option: Optional[bool] = None,
    ) -> PollingOperation[None]:
        """An asynchronous version of :func:`grant_future_privileges`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """  # noqa: D401
        grant = Grant(
            privileges=privileges,
            securable_type=securable_type,
            containing_scope=containing_scope,
            grant_option=grant_option,
        )
        future = self.collection._api.grant_future_privileges(self.name, grant, async_req=True)
        return PollingOperations.empty(future)

    @api_telemetry
    def revoke_role(self, role_type: str, role: Securable) -> None:
        """
        Revoke a role from this role.

        Parameters
        __________
        role_type: str
            The type of role which would be revoked.
        role: Securable
            The role which would be revoked.

        Examples
        ________
        Using a role reference to revoke a role from it:

        >>> reference_role.revoke("role", Securable(name="test_role"))
        """
        grant = Grant(securable_type=role_type, securable=role)
        self.collection._api.revoke_grants(self.name, grant, async_req=False)

    @api_telemetry
    def revoke_role_async(self, role_type: str, role: Securable) -> PollingOperation[None]:
        """An asynchronous version of :func:`revoke_role`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """  # noqa: D401
        grant = Grant(securable_type=role_type, securable=role)
        future = self.collection._api.revoke_grants(self.name, grant, async_req=True)
        return PollingOperations.empty(future)

    @api_telemetry
    def revoke_privileges(
        self,
        privileges: list[str],
        securable_type: str,
        securable: Optional[Securable] = None,
        mode: Optional[DeleteMode] = None,
    ) -> None:
        """
        Revoke privileges on a securable from this role.

        Parameters
        __________
        privileges: list[str]
            The list of privileges to be revoked.
        securable_type: str
            The type of securable on which the privileges would be revoked.
        securable: Securable, optional
            The securable on which the privileges would be revoked. Default is None.
        mode: DeleteMode, optional
            One of the following enum values.

            ``DeleteMode.restrict``: If the privilege being revoked has been re-granted to another role,
            the REVOKE command fails.

            ``DeleteMode.cascade``: If the privilege being revoked has been re-granted, the REVOKE command
            recursively revokes these dependent grants. If the same privilege on an object has been granted
            to the target role by a different grantor (parallel grant), that grant is not affected and the
            target role retains the privilege.

            Default is None which is equivalent to ``DeleteMode.restrict``.

        Examples
        ________
        Using a role reference to revoke privileges from it:

        >>> reference_role.revoke_privileges(["CREATE", "USAGE"], "database", Securable(database="test_db"))
        """
        grant = Grant(privileges=privileges, securable_type=securable_type, securable=securable)
        self.collection._api.revoke_grants(self.name, grant, mode, async_req=False)

    @api_telemetry
    def revoke_privileges_async(
        self,
        privileges: list[str],
        securable_type: str,
        securable: Optional[Securable] = None,
        mode: Optional[DeleteMode] = None,
    ) -> PollingOperation[None]:
        """An asynchronous version of :func:`revoke_privileges`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """  # noqa: D401
        grant = Grant(privileges=privileges, securable_type=securable_type, securable=securable)
        future = self.collection._api.revoke_grants(self.name, grant, mode, async_req=True)
        return PollingOperations.empty(future)

    @api_telemetry
    def revoke_privileges_on_all(
        self,
        privileges: list[str],
        securable_type: str,
        containing_scope: ContainingScope,
        mode: Optional[DeleteMode] = None,
    ) -> None:
        """
        Revoke privileges on all securables matching the criteria from this role.

        Parameters
        __________
        privileges: list[str]
            The list of privileges to be revoked.
        securable_type: str
            The type of securable on which the privileges would be revoked.
        containing_scope: ContainingScope
            The criteria to match the securables.
        mode: DeleteMode, optional
            One of the following enum values.

            ``DeleteMode.restrict``: If the privilege being revoked has been re-granted to another role,
            the REVOKE command fails.

            ``DeleteMode.cascade``: If the privilege being revoked has been re-granted, the REVOKE command
            recursively revokes these dependent grants. If the same privilege on an object has been granted
            to the target role by a different grantor (parallel grant), that grant is not affected and the
            target role retains the privilege.

            Default is None which is equivalent to ``DeleteMode.restrict``.

        Examples
        ________
        Using a role reference to revoke privileges on all schemas in a database from it:

        >>> reference_role.revoke_privileges_on_all(
        ...     ["CREATE", "USAGE"], "schema", ContainingScope(database="test_db")
        ... )
        """
        grant = Grant(privileges=privileges, securable_type=securable_type, containing_scope=containing_scope)
        self.collection._api.revoke_grants(self.name, grant, mode, async_req=False)

    @api_telemetry
    def revoke_privileges_on_all_async(
        self,
        privileges: list[str],
        securable_type: str,
        containing_scope: ContainingScope,
        mode: Optional[DeleteMode] = None,
    ) -> PollingOperation[None]:
        """An asynchronous version of :func:`revoke_privileges_on_all`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """  # noqa: D401
        grant = Grant(privileges=privileges, securable_type=securable_type, containing_scope=containing_scope)
        future = self.collection._api.revoke_grants(self.name, grant, mode, async_req=True)
        return PollingOperations.empty(future)

    @api_telemetry
    def revoke_future_privileges(
        self,
        privileges: list[str],
        securable_type: str,
        containing_scope: ContainingScope,
        mode: Optional[DeleteMode] = None,
    ) -> None:
        """
        Revoke privileges on all future securables matching the criteria from this role.

        Parameters
        __________
        privileges: list[str]
            The list of privileges to be revoked.
        securable_type: str
            The type of securable on which the privileges would be revoked.
        containing_scope: ContainingScope
            The criteria to match the securables.
        mode: DeleteMode, optional
            One of the following enum values.

            ``DeleteMode.restrict``: If the privilege being revoked has been re-granted to another role,
            the REVOKE command fails.

            ``DeleteMode.cascade``: If the privilege being revoked has been re-granted, the REVOKE command
            recursively revokes these dependent grants. If the same privilege on an object has been granted
            to the target role by a different grantor (parallel grant), that grant is not affected and the
            target role retains the privilege.

            Default is None which is equivalent to ``DeleteMode.restrict``.

        Examples
        ________
        Using a role reference to revoke privileges on all future schemas in a database from it:

        >>> reference_role.revoke_future_privileges(
        ...     ["CREATE", "USAGE"], "schema", ContainingScope(database="test_db")
        ... )
        """
        grant = Grant(privileges=privileges, securable_type=securable_type, containing_scope=containing_scope)
        self.collection._api.revoke_future_grants(self.name, grant, mode, async_req=False)

    @api_telemetry
    def revoke_future_privileges_async(
        self,
        privileges: list[str],
        securable_type: str,
        containing_scope: ContainingScope,
        mode: Optional[DeleteMode] = None,
    ) -> PollingOperation[None]:
        """An asynchronous version of :func:`revoke_future_privileges`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """  # noqa: D401
        grant = Grant(privileges=privileges, securable_type=securable_type, containing_scope=containing_scope)
        future = self.collection._api.revoke_future_grants(self.name, grant, mode, async_req=True)
        return PollingOperations.empty(future)

    @api_telemetry
    def revoke_grant_option_for_privileges(
        self,
        privileges: list[str],
        securable_type: str,
        securable: Optional[Securable] = None,
        mode: Optional[DeleteMode] = None,
    ) -> None:
        """
        Revoke grant option for privileges on a securable from this role.

        Parameters
        __________
        privileges: list[str]
            The list of privileges to be revoked.
        securable_type: str
            The type of securable on which the privileges would be revoked.
        securable: Securable, optional
            The securable on which the privileges would be revoked. Default is None.
        mode: DeleteMode, optional
            One of the following enum values.

            ``DeleteMode.restrict``: If the privilege being revoked has been re-granted to another role,
            the REVOKE command fails.

            ``DeleteMode.cascade``: If the privilege being revoked has been re-granted, the REVOKE command
            recursively revokes these dependent grants. If the same privilege on an object has been granted
            to the target role by a different grantor (parallel grant), that grant is not affected and the
            target role retains the privilege.

            Default is None which is equivalent to ``DeleteMode.restrict``.

        Examples
        ________
        Using a role reference to revoke grant option for privileges from it:

        >>> reference_role.revoke_grant_option_for_privileges(
        ...     ["CREATE", "USAGE"], "database", Securable(database="test_db")
        ... )
        """
        grant = Grant(privileges=privileges, securable_type=securable_type, securable=securable, grant_option=True)
        self.collection._api.revoke_grants(self.name, grant, mode, async_req=False)

    @api_telemetry
    def revoke_grant_option_for_privileges_async(
        self,
        privileges: list[str],
        securable_type: str,
        securable: Optional[Securable] = None,
        mode: Optional[DeleteMode] = None,
    ) -> PollingOperation[None]:
        """An asynchronous version of :func:`revoke_grant_option_for_privileges`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """  # noqa: D401
        grant = Grant(privileges=privileges, securable_type=securable_type, securable=securable, grant_option=True)
        future = self.collection._api.revoke_grants(self.name, grant, mode, async_req=True)
        return PollingOperations.empty(future)

    @api_telemetry
    def revoke_grant_option_for_privileges_on_all(
        self,
        privileges: list[str],
        securable_type: str,
        containing_scope: ContainingScope,
        mode: Optional[DeleteMode] = None,
    ) -> None:
        """
        Revoke grant option for privileges on all securables matching the criteria from this role.

        Parameters
        __________
        privileges: list[str]
            The list of privileges to be revoked.
        securable_type: str
            The type of securable on which the privileges would be revoked.
        containing_scope: ContainingScope
            The criteria to match the securables.
        mode: DeleteMode, optional
            One of the following enum values.

            ``DeleteMode.restrict``: If the privilege being revoked has been re-granted to another role,
            the REVOKE command fails.

            ``DeleteMode.cascade``: If the privilege being revoked has been re-granted, the REVOKE command
            recursively revokes these dependent grants. If the same privilege on an object has been granted
            to the target role by a different grantor (parallel grant), that grant is not affected and the
            target role retains the privilege.

            Default is None which is equivalent to ``DeleteMode.restrict``.

        Examples
        ________
        Using a role reference to revoke grant option for privileges on all schemas in a database from it:

        >>> reference_role.revoke_grant_option_for_privileges_on_all(
        ...     ["CREATE", "USAGE"], "schema", ContainingScope(database="test_db")
        ... )
        """
        grant = Grant(
            privileges=privileges, securable_type=securable_type, containing_scope=containing_scope, grant_option=True
        )
        self.collection._api.revoke_grants(self.name, grant, mode, async_req=False)

    @api_telemetry
    def revoke_grant_option_for_privileges_on_all_async(
        self,
        privileges: list[str],
        securable_type: str,
        containing_scope: ContainingScope,
        mode: Optional[DeleteMode] = None,
    ) -> PollingOperation[None]:
        """An asynchronous version of :func:`revoke_grant_option_for_privileges_on_all`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """  # noqa: D401
        grant = Grant(
            privileges=privileges, securable_type=securable_type, containing_scope=containing_scope, grant_option=True
        )
        future = self.collection._api.revoke_grants(self.name, grant, mode, async_req=True)
        return PollingOperations.empty(future)

    @api_telemetry
    def revoke_grant_option_for_future_privileges(
        self,
        privileges: list[str],
        securable_type: str,
        containing_scope: ContainingScope,
        mode: Optional[DeleteMode] = None,
    ) -> None:
        """
        Revoke grant option for privileges on all future securables matching the criteria from this role.

        Parameters
        __________
        privileges: list[str]
            The list of privileges to be revoked.
        securable_type: str
            The type of securable on which the privileges would be revoked.
        containing_scope: ContainingScope
            The criteria to match the securables.
        mode: DeleteMode, optional
            One of the following enum values.

            ``DeleteMode.restrict``: If the privilege being revoked has been re-granted to another role,
            the REVOKE command fails.

            ``DeleteMode.cascade``: If the privilege being revoked has been re-granted, the REVOKE command
            recursively revokes these dependent grants. If the same privilege on an object has been granted
            to the target role by a different grantor (parallel grant), that grant is not affected and the
            target role retains the privilege.

            Default is None which is equivalent to ``DeleteMode.restrict``.

        Examples
        ________
        Using a role reference to revoke grant option for privileges on all future schemas in a database from it:

        >>> reference_role.revoke_grant_option_for_future_privileges(
        ...     ["CREATE", "USAGE"], "schema", ContainingScope(database="test_db")
        ... )
        """
        grant = Grant(
            privileges=privileges, securable_type=securable_type, containing_scope=containing_scope, grant_option=True
        )
        self.collection._api.revoke_future_grants(self.name, grant, mode, async_req=False)

    @api_telemetry
    def revoke_grant_option_for_future_privileges_async(
        self,
        privileges: list[str],
        securable_type: str,
        containing_scope: ContainingScope,
        mode: Optional[DeleteMode] = None,
    ) -> PollingOperation[None]:
        """An asynchronous version of :func:`revoke_grant_option_for_future_privileges`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """  # noqa: D401
        grant = Grant(
            privileges=privileges, securable_type=securable_type, containing_scope=containing_scope, grant_option=True
        )
        future = self.collection._api.revoke_future_grants(self.name, grant, mode, async_req=True)
        return PollingOperations.empty(future)

    @api_telemetry
    def iter_grants_of(self, show_limit: Optional[int] = None) -> Iterator[GrantOf]:
        """List grants of this role.

        Lists all users and roles to which the role has been granted.

        Parameters
        __________
        show_limit: int, optional
            Limit of the maximum number of rows returned by iter(). The default is ``None``, which behaves equivalently
            to show_limit=10000. This value must be between ``1`` and ``10000``.

        Examples
        ________
        Using a role reference to list grants of it:

        >>> reference_role.iter_grants_of()
        """
        grants_list = self.collection._api.list_grants_of(self.name, show_limit, async_req=False)
        return iter(grants_list)

    @api_telemetry
    def iter_grants_of_async(self, show_limit: Optional[int] = None) -> PollingOperation[Iterator[GrantOf]]:
        """An asynchronous version of :func:`iter_grants_of`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """  # noqa: D401
        future = self.collection._api.list_grants_of(self.name, show_limit, async_req=True)
        return PollingOperations.iterator(future)

    @api_telemetry
    def iter_grants_on(self, show_limit: Optional[int] = None) -> Iterator[GrantOn]:
        """List grants on this role.

        Lists all privileges that have been granted on the role.

        Parameters
        __________
        show_limit: int, optional
            Limit of the maximum number of rows returned by iter(). The default is ``None``, which behaves equivalently
            to show_limit=10000. This value must be between ``1`` and ``10000``.

        Examples
        ________
        Using a role reference to list grants on it:

        >>> reference_role.iter_grants_on()
        """
        grants_list = self.collection._api.list_grants_on(self.name, show_limit, async_req=False)
        return iter(grants_list)

    @api_telemetry
    def iter_grants_on_async(self, show_limit: Optional[int] = None) -> PollingOperation[Iterator[GrantOn]]:
        """An asynchronous version of :func:`iter_grants_on`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """  # noqa: D401
        future = self.collection._api.list_grants_on(self.name, show_limit, async_req=True)
        return PollingOperations.iterator(future)

    @api_telemetry
    def iter_grants_to(self, show_limit: Optional[int] = None) -> Iterator[Grant]:
        """List grants to this role.

        Lists all privileges and roles granted to the role.

        Parameters
        __________
        show_limit: int, optional
            Limit of the maximum number of rows returned by iter(). The default is ``None``, which behaves equivalently
            to show_limit=10000. This value must be between ``1`` and ``10000``.

        Examples
        ________
        Using a role reference to list grants to it:

        >>> reference_role.iter_grants_to()
        """
        grants_list = self.collection._api.list_grants(self.name, show_limit, async_req=False)
        return iter(grants_list)

    @api_telemetry
    def iter_grants_to_async(self, show_limit: Optional[int] = None) -> PollingOperation[Iterator[Grant]]:
        """An asynchronous version of :func:`iter_grants_to`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """  # noqa: D401
        future = self.collection._api.list_grants(self.name, show_limit, async_req=True)
        return PollingOperations.iterator(future)

    @api_telemetry
    def iter_future_grants_to(self, show_limit: Optional[int] = None) -> Iterator[Grant]:
        """List future grants to this role.

        Lists all privileges on new (i.e. future) objects of a specified type in a database
        or schema granted to the role.

        Parameters
        __________
        show_limit: int, optional
            Limit of the maximum number of rows returned by iter(). The default is ``None``, which behaves equivalently
            to show_limit=10000. This value must be between ``1`` and ``10000``.

        Examples
        ________
        Using a role reference to list future grants to it:

        >>> reference_role.iter_future_grants_to()
        """
        grants_list = self.collection._api.list_future_grants(self.name, show_limit, async_req=False)
        return iter(grants_list)

    @api_telemetry
    def iter_future_grants_to_async(self, show_limit: Optional[int] = None) -> PollingOperation[Iterator[Grant]]:
        """An asynchronous version of :func:`iter_future_grants_to`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """  # noqa: D401
        future = self.collection._api.list_future_grants(self.name, show_limit, async_req=True)
        return PollingOperations.iterator(future)

    @api_telemetry
    def set_tags(self, tags: dict[TagResource, TagValue], if_exists: Optional[bool] = None) -> None:
        """Set tags on a role.

        Parameters
        __________
        tags: dict[TagResource, TagValue]
             (required)
        if_exists: bool
             Parameter that specifies how to handle the request for a resource that does not exist: - `true`:
             The endpoint does not throw an error if the resource does not exist. It returns a 200 success response,
             but does not take any action on the resource. - `false`: The endpoint throws an error if the resource
             doesn't exist.
        """
        self.collection._api.set_tags(
            name=self.name,
            tag_assignment=[
                tag_tuple_to_tag_assignment(TagAssignment, tag_resource, tag_value)
                for [tag_resource, tag_value] in tags.items()
            ],
            if_exists=if_exists,
        )

    @api_telemetry
    def set_tags_async(
        self, tags: dict[TagResource, TagValue], if_exists: Optional[bool] = None
    ) -> PollingOperation[None]:
        """An asynchronous version of :func:`set_tags`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """  # noqa: D401
        future = self.collection._api.set_tags(
            name=self.name,
            tag_assignment=[
                tag_tuple_to_tag_assignment(TagAssignment, tag_resource, tag_value)
                for [tag_resource, tag_value] in tags.items()
            ],
            if_exists=if_exists,
            async_req=True,
        )
        return PollingOperations.empty(future)

    @api_telemetry
    def unset_tags(self, tag_resources: set[TagResource], if_exists: Optional[bool] = None) -> None:
        """Unset tags from a role.

        Parameters
        __________
        tag_resources: set[TagResource]
             (required)
        if_exists: bool
             Parameter that specifies how to handle the request for a resource that does not exist: - `true`:
             The endpoint does not throw an error if the resource does not exist. It returns a 200 success response,
             but does not take any action on the resource. - `false`: The endpoint throws an error if the resource
             doesn't exist.
        """
        self.collection._api.unset_tags(
            name=self.name,
            tag_reference=[tag_resource_to_tag_reference(TagReference, tag_resource) for tag_resource in tag_resources],
            if_exists=if_exists,
        )

    @api_telemetry
    def unset_tags_async(
        self, tag_resources: set[TagResource], if_exists: Optional[bool] = None
    ) -> PollingOperation[None]:
        """An asynchronous version of :func:`unset_tags`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """  # noqa: D401
        future = self.collection._api.unset_tags(
            name=self.name,
            tag_reference=[tag_resource_to_tag_reference(TagReference, tag_resource) for tag_resource in tag_resources],
            if_exists=if_exists,
            async_req=True,
        )
        return PollingOperations.empty(future)

    @api_telemetry
    def get_tags(self, with_lineage: Optional[bool] = None) -> dict[TagResource, TagValue]:
        """Get the tag assignments for a role.

        Returns all tags assigned to a role. This operation requires an active warehouse.

        Parameters
        __________
        with_lineage: bool, optional
            Parameter that specifies whether tag assignments inherited by the object from its ancestors in securable
            object hierarchy should be returned as well: - `true`: All tags assigned to this object should be returned,
            inheritance included. - `false`: Only tags explicitly assigned to this object should be returned.
        """
        tag_assignments = self.collection._api.get_tags(
            name=self.name,
            with_lineage=with_lineage,
        )
        return dict(tag_assignment_to_tag_tuple(ta, self.root) for ta in tag_assignments)

    @api_telemetry
    def get_tags_async(self, with_lineage: Optional[bool] = None) -> PollingOperation[dict[TagResource, TagValue]]:
        """An asynchronous version of :func:`get_tags`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """  # noqa: D401
        future = self.collection._api.get_tags(name=self.name, with_lineage=with_lineage, async_req=True)
        return PollingOperation(
            future, lambda tag_assignments: dict(tag_assignment_to_tag_tuple(ta, self.root) for ta in tag_assignments)
        )



from unittest import mock

import pytest

from snowflake.core import PollingOperation
from snowflake.core.secret import PasswordSecret, SecretResource

from ...utils import BASE_URL, extra_params, mock_http_response


@pytest.fixture
def secrets(schema):
    return schema.secrets


@pytest.fixture
def secret(secrets):
    return secrets["my_secret"]


API_CLIENT_REQUEST = "snowflake.core._generated.api_client.ApiClient.request"
SECRET = PasswordSecret(
    name="my_secret",
    username="snowman",
    password="test",
    comment="Test secret",
)


def test_create_secret(fake_root, secrets):
    kwargs = extra_params(
        query_params=[],
        body={
            "name": "my_secret",
            "username": "snowman",
            "password": "test",
            "type": "PASSWORD",
            "comment": "Test secret",
        },
    )
    args = (fake_root, "POST", BASE_URL + "/databases/my_db/schemas/my_schema/secrets")

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        s_res = secrets.create(SECRET)
        assert isinstance(s_res, SecretResource)
        assert s_res.name == "my_secret"
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = secrets.create_async(SECRET)
        assert isinstance(op, PollingOperation)
        s_res = op.result()
        assert s_res.name == "my_secret"
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_iter_secret(fake_root, secrets):
    args = (fake_root, "GET", BASE_URL + "/databases/my_db/schemas/my_schema/secrets")
    kwargs = extra_params()

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response()
        secrets.iter()
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response()
        op = secrets.iter_async()
        assert isinstance(op, PollingOperation)
        it = op.result()
        assert list(it) == []
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_iter_show_limit_deprecation_warning(fake_root, secrets):
    args = (
        fake_root,
        "GET",
        BASE_URL + "/databases/my_db/schemas/my_schema/secrets?showLimit=10",
    )
    kwargs = extra_params(query_params=[("showLimit", 10)])

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response()
        with pytest.warns(DeprecationWarning, match="'show_limit' is deprecated, use 'limit' instead"):
            list(secrets.iter(show_limit=10))
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response()
        with pytest.warns(DeprecationWarning, match="'show_limit' is deprecated, use 'limit' instead"):
            secrets.iter_async(show_limit=10).result()
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_iter_limit_and_show_limit_conflict(secrets):
    with pytest.raises(ValueError, match="Cannot specify both 'limit' and 'show_limit'"):
        list(secrets.iter(limit=10, show_limit=5))

    with pytest.raises(ValueError, match="Cannot specify both 'limit' and 'show_limit'"):
        secrets.iter_async(limit=10, show_limit=5).result()


def test_fetch_secret(fake_root, secret):
    args = (fake_root, "GET", BASE_URL + "/databases/my_db/schemas/my_schema/secrets/my_secret")
    kwargs = extra_params()

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response(SECRET.to_json())
        s = secret.fetch()
        assert s.to_dict() == SECRET.to_dict()
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response(SECRET.to_json())
        op = secret.fetch_async()
        assert isinstance(op, PollingOperation)
        s = op.result()
        assert s.to_dict() == SECRET.to_dict()
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_drop_secret(fake_root, secret):
    args = (fake_root, "DELETE", BASE_URL + "/databases/my_db/schemas/my_schema/secrets/my_secret")
    kwargs = extra_params()

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        secret.drop()
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = secret.drop_async()
        assert isinstance(op, PollingOperation)
        op.result()
    mocked_request.assert_called_once_with(*args, **kwargs)



"""
Snowflake Services API.

The Snowflake Services API is a REST API that you can use to access, update, and perform certain actions on Services resource in a Snowflake database.  # noqa: E501

The version of the OpenAPI document: 0.0.1
Contact: support@snowflake.com
Generated by: https://openapi-generator.tech

Do not edit this file manually.
"""

from __future__ import annotations

import json
import pprint
import re  # noqa: F401

from datetime import datetime
from typing import Any, Optional

from pydantic import BaseModel, ConfigDict, StrictStr


class GrantOf(BaseModel):
    """A model object representing the GrantOf resource.

    Constructs an object of type GrantOf with the provided properties.

    Parameters
    __________
    created_on : datetime, optional
        Date and time when the grant was created — **Read-only:** *any user-provided value will be ignored.*
    role : str, optional
        The name of the service role — **Read-only:** *any user-provided value will be ignored.*
    granted_to : str, optional
        The type of the grantee, can be USER or ROLE — **Read-only:** *any user-provided value will be ignored.*
    grantee_name : str, optional
        The name of the grantee — **Read-only:** *any user-provided value will be ignored.*
    granted_by : str, optional
        The name of role that granted the service role to the grantee — **Read-only:** *any user-provided value will be ignored.*
    """

    created_on: Optional[datetime] = None

    role: Optional[StrictStr] = None

    granted_to: Optional[StrictStr] = None

    grantee_name: Optional[StrictStr] = None

    granted_by: Optional[StrictStr] = None

    __properties = ["created_on", "role", "granted_to", "grantee_name", "granted_by"]

    model_config = ConfigDict(
        validate_by_name=True,
        validate_assignment=True,
    )

    def to_str(self) -> str:
        """Returns the string representation of the model using alias."""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias."""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> GrantOf:
        """Create an instance of GrantOf from a JSON string."""
        return cls.from_dict(json.loads(json_str))

    def to_dict(
        self,
        hide_readonly_properties: bool = False,
    ) -> dict[str, Any]:
        """Returns the dictionary representation of the model using alias."""
        exclude_properties = set()

        if hide_readonly_properties:
            exclude_properties.update(
                {
                    "created_on",
                    "role",
                    "granted_to",
                    "grantee_name",
                    "granted_by",
                }
            )

        _dict = self.model_dump(serialize_as_any=True, by_alias=True, exclude=exclude_properties, exclude_none=True)

        return _dict

    def to_dict_without_readonly_properties(self) -> dict[str, Any]:
        """Return the dictionary representation of the model without readonly properties."""
        return self.to_dict(hide_readonly_properties=True)

    @classmethod
    def from_dict(cls, obj: dict) -> GrantOf:
        """Create an instance of GrantOf from a dict."""
        if obj is None:
            return None

        if type(obj) is not dict:
            return GrantOf.model_validate(obj)

        _obj = GrantOf.model_validate(
            {
                "created_on": obj.get("created_on"),
                "role": obj.get("role"),
                "granted_to": obj.get("granted_to"),
                "grantee_name": obj.get("grantee_name"),
                "granted_by": obj.get("granted_by"),
            }
        )

        return _obj


class GrantOfModel:
    def __init__(
        self,  # optional properties
        created_on: Optional[datetime] = None,
        role: Optional[str] = None,
        granted_to: Optional[str] = None,
        grantee_name: Optional[str] = None,
        granted_by: Optional[str] = None,
    ):
        """A model object representing the GrantOf resource.

        Constructs an object of type GrantOf with the provided properties.

        Parameters
        __________
        created_on : datetime, optional
            Date and time when the grant was created
        role : str, optional
            The name of the service role
        granted_to : str, optional
            The type of the grantee, can be USER or ROLE
        grantee_name : str, optional
            The name of the grantee
        granted_by : str, optional
            The name of role that granted the service role to the grantee
        """
        self.created_on = created_on
        self.role = role
        self.granted_to = granted_to
        self.grantee_name = grantee_name
        self.granted_by = granted_by

    __properties = ["created_on", "role", "granted_to", "grantee_name", "granted_by"]

    def __repr__(self) -> str:
        return repr(self._to_model())

    def _to_model(self):
        return GrantOf(
            created_on=self.created_on,
            role=self.role,
            granted_to=self.granted_to,
            grantee_name=self.grantee_name,
            granted_by=self.granted_by,
        )

    @classmethod
    def _from_model(cls, model) -> GrantOfModel:
        return GrantOfModel(
            created_on=model.created_on,
            role=model.role,
            granted_to=model.granted_to,
            grantee_name=model.grantee_name,
            granted_by=model.granted_by,
        )

    def to_dict(self):
        """Create a dictionary of the properties from a GrantOf.

        This method constructs a dictionary with the key-value entries corresponding to the properties of the GrantOf object.

        Returns
        _______
        dict
            A dictionary object created using the input model.
        """
        return self._to_model().to_dict()

    @classmethod
    def from_dict(cls, obj: dict) -> GrantOfModel:
        """Create an instance of GrantOf from a dict.

        This method constructs a GrantOf object from a dictionary with the key-value pairs of its properties.

        Parameters
        ----------
        obj : dict
            A dictionary whose keys and values correspond to the properties of the resource object.

        Returns
        _______
        GrantOf
            A GrantOf object created using the input dictionary; this will fail if the required properties are missing.
        """
        return cls._from_model(GrantOf.from_dict(obj))


GrantOf._model_class = GrantOfModel



"""Manages Snowflake streamlit."""

from ._generated.models import (
    AddLiveVersionStreamlitRequest,
    AddVersionFromGitStreamlitRequest,
    AddVersionStreamlitRequest,
    AddVersionStreamlitRequestVersion,
    CommitStreamlitRequest,
    GitPushWithUsernamePassword,
    Streamlit,
    StreamlitPushOptions,
    StreamlitVersionForGit,
)
from ._streamlit import StreamlitCollection, StreamlitResource


__all__ = [
    "StreamlitResource",
    "StreamlitCollection",
    "AddLiveVersionStreamlitRequest",
    "AddVersionFromGitStreamlitRequest",
    "AddVersionStreamlitRequest",
    "AddVersionStreamlitRequestVersion",
    "CommitStreamlitRequest",
    "GitPushWithUsernamePassword",
    "Streamlit",
    "StreamlitPushOptions",
    "StreamlitVersionForGit",
]



#
# Copyright (c) 2012-2023 Snowflake Computing Inc. All rights reserved.
#

from io import BytesIO
from textwrap import dedent

import pytest

from snowflake.core.role import Role, Securable
from snowflake.core.service import GrantOf, Service, ServiceRoleGrantTo, ServiceSpecStageFile
from tests.utils import random_string


pytestmark = [pytest.mark.skip_gov]


@pytest.mark.min_sf_ver("8.39.0")
@pytest.mark.usefixtures("qa_mode_enabled")
def test_service_roles(session, database, schema, services, roles, imagerepo, shared_compute_pool):
    role = None
    svc = None

    try:
        # Creating a custom role
        role_name = random_string(4, "test_role_")
        test_role = Role(name=role_name, comment="test role")
        role = roles.create(test_role)

        # Create a service
        service_name = random_string(5, "test_service_")
        stage_name = random_string(5, "test_stage_")
        session.sql(f"create temp stage {stage_name};").collect()
        spec_file = "spec.yaml"
        stage_file = f"@{stage_name}"
        spec = f"{stage_file}/{spec_file}"
        session.file.put_stream(
            BytesIO(
                dedent(f"""
                        spec:
                          containers:
                          - name: main-container
                            image: {imagerepo}/hello-world:latest
                          - name: other-container
                            image: {imagerepo}/hello-world:latest
                          endpoints:
                          - name: echo
                            port: 8080
                          - name: anotherecho
                            port: 8085
                        serviceRoles:
                        - name: role1
                          endpoints:
                          - echo
                        """).encode()
            ),
            spec,
        )

        test_service = Service(
            name=service_name,
            compute_pool=shared_compute_pool,
            spec=ServiceSpecStageFile(stage=stage_name, spec_file=spec_file),
            min_instances=1,
            max_instances=1,
        )

        svc = services.create(test_service)

        # Grant service role to the test role
        assert len(list(role.iter_grants_to())) == 0
        role.grant_role(
            role_type="SERVICE ROLE",
            role=Securable(database=database.name, schema=schema.name, service=service_name, name="role1"),
        )
        assert len(list(role.iter_grants_to())) == 1

        # Verify the output of the iter_grants_of_service_role command
        expected_grants_of = [
            GrantOf.from_dict(
                {
                    "created_on": "2012-08-01T07:00:00.000+00:00",
                    "role": f"{database.name.upper()}.{schema.name.upper()}.{service_name.upper()}.ROLE1",
                    "granted_to": "ROLE",
                    "grantee_name": "ACCOUNTADMIN",
                    "granted_by": "",
                }
            ),
            GrantOf.from_dict(
                {
                    "created_on": "2012-08-01T07:00:00.000+00:00",
                    "role": f"{database.name.upper()}.{schema.name.upper()}.{service_name.upper()}.ROLE1",
                    "granted_to": "ROLE",
                    "grantee_name": role_name.upper(),
                    "granted_by": "ACCOUNTADMIN",
                }
            ),
        ]

        actual_grants_of = list(svc.iter_grants_of_service_role("role1"))
        assert actual_grants_of == expected_grants_of

        # Verify the output of the iter_grants_to_service_role command
        expected_grants_to = [
            ServiceRoleGrantTo.from_dict(
                {
                    "created_on": "2012-08-01T07:00:00.000+00:00",
                    "privilege": "USAGE",
                    "granted_on": "SERVICE_ENDPOINT",
                    "name": f"{database.name.upper()}.{schema.name.upper()}.{service_name.upper()}!echo",
                    "granted_to": "SERVICE ROLE",
                    "grantee_name": "ROLE1",
                }
            )
        ]

        actual_grants_to = list(svc.iter_grants_to_service_role("role1"))
        assert actual_grants_to == expected_grants_to

        # Revoke service role from account role
        role.revoke_role(
            role_type="SERVICE ROLE",
            role=Securable(database=database.name, schema=schema.name, service=service_name, name="role1"),
        )
        assert len(list(role.iter_grants_to())) == 0
    finally:
        if role is not None:
            role.drop()

        if svc is not None:
            svc.drop()



from collections.abc import Iterator
from concurrent.futures import ThreadPoolExecutor

import pytest

from snowflake.core._operation import PollingOperation, PollingOperations


@pytest.fixture()
def pool():
    pool = ThreadPoolExecutor(max_workers=1)
    yield pool
    pool.shutdown()


def test_return_result(pool):
    f = pool.submit(lambda: 5)
    f2 = PollingOperation(f, lambda num: num * 2)
    assert f2.result() == 10


def test_delegate_throws(pool):
    def throw():
        raise ValueError("Oops")

    f = pool.submit(throw)
    f2 = PollingOperation(f, lambda x: x)
    with pytest.raises(ValueError, match="Oops"):
        f2.result()


def test_fn_throws(pool):
    def throw(_):
        raise ValueError("Oops")

    f = pool.submit(lambda: 5)
    f2 = PollingOperation(f, throw)
    with pytest.raises(ValueError, match="Oops"):
        f2.result()


def test_cancel_when_not_started(pool, event):
    pool.submit(lambda: event.wait())
    f = pool.submit(lambda: event.wait())
    f2 = PollingOperation(f, lambda x: x)
    assert f2.cancel()
    assert f2.cancelled()


def test_cancel_when_already_started(pool, event):
    f = pool.submit(lambda: event.wait())
    f2 = PollingOperation(f, lambda x: x)
    assert not f2.cancel()
    assert not f2.cancelled()


def test_empty(pool):
    f = pool.submit(lambda: 5)
    op = PollingOperations.empty(f)
    assert op.result() is None


def test_identity(pool):
    f = pool.submit(lambda: 5)
    op = PollingOperations.identity(f)
    assert op.result() == 5


def test_iterator(pool):
    f = pool.submit(lambda: [1, 2, 3])
    op = PollingOperations.iterator(f)
    it = op.result()
    assert isinstance(it, Iterator)
    assert list(it) == [1, 2, 3]



"""
Snowflake View API.

The Snowflake View API is a REST API that you can use to access, update, and perform certain actions on View resource in a Snowflake database.  # noqa: E501

The version of the OpenAPI document: 0.0.1
Contact: support@snowflake.com
Generated by: https://openapi-generator.tech

Do not edit this file manually.
"""

import warnings

from typing import TYPE_CHECKING, Iterator, Optional, TypeVar, Union

from snowflake.core._common import CreateMode, SchemaObjectCollectionParent, SchemaObjectReferenceMixin
from snowflake.core._generated.api_client import StoredProcApiClient
from snowflake.core._internal.telemetry import api_telemetry
from snowflake.core._operation import PollingOperation, PollingOperations
from snowflake.core._utils import get_function_name_with_args, replace_function_name_in_name_with_args
from snowflake.core.tag import TagResource, TagValue
from snowflake.core.view._generated.models.tag_assignment import TagAssignment
from snowflake.core.view._generated.models.tag_reference import TagReference
from snowflake.core.view._generated.models.view import View

from .view_api import ViewApi


if TYPE_CHECKING:
    from snowflake.core.schema import SchemaResource


T = TypeVar("T")


class ViewCollectionBase(SchemaObjectCollectionParent["ViewResource"]):
    _identifier_requires_args: bool = False

    def __init__(self, schema: "SchemaResource", resource_class: type[T]) -> None:
        super().__init__(schema, resource_class)
        self._api = ViewApi(
            root=self.root, resource_class=self._ref_class, sproc_client=StoredProcApiClient(root=self.root)
        )

    @api_telemetry
    def create(
        self, view: View, mode: Optional[Union[CreateMode, str]] = None, copy_grants: Optional[bool] = None
    ) -> "ViewResource":  # noqa: F821
        """Create a view.

        Parameters
        __________
        view: View
             (required)
        mode: Union[CreateMode, str]
             Parameter allowing support for different modes of resource creation. Possible values include: - `errorIfExists`: Throws an error if you try to create a resource that already exists. - `orReplace`: Automatically replaces the existing resource with the current one. - `ifNotExists`: Creates a new resource when an alter is requested for a non-existent resource.
        copy_grants: bool
             Parameter to enable copy grants when creating the object.
        """
        if isinstance(mode, str):
            mode = CreateMode[mode].value

        self._api.create_view(
            self.database.name,
            self.schema.name,
            view=view,
            create_mode=mode,
            copy_grants=copy_grants,
            async_req=False,
        )
        if self._identifier_requires_args:
            return self._ref_class(get_function_name_with_args(view), self)
        return self._ref_class(view.name, self)

    @api_telemetry
    def create_async(
        self, view: View, mode: Optional[Union[CreateMode, str]] = None, copy_grants: Optional[bool] = None
    ) -> PollingOperation["ViewResource"]:  # noqa: F821
        """An asynchronous version of :func:`create`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """
        if isinstance(mode, str):
            mode = CreateMode[mode].value

        future = self._api.create_view(
            self.database.name,
            self.schema.name,
            view=view,
            create_mode=mode,
            copy_grants=copy_grants,
            async_req=True,
        )
        if self._identifier_requires_args:
            return PollingOperation(future, lambda _: self._ref_class(get_function_name_with_args(view), self))
        return PollingOperation(future, lambda _: self._ref_class(view if isinstance(view, str) else view.name, self))

    @api_telemetry
    def iter(
        self,
        *,
        like: Optional[str] = None,
        starts_with: Optional[str] = None,
        limit: Optional[int] = None,
        from_name: Optional[str] = None,
        deep: Optional[bool] = None,
        show_limit: Optional[int] = None,
    ) -> Iterator[View]:
        """List views.

        Parameters
        __________
        like: str
             Parameter to filter the command output by resource name. Uses case-insensitive pattern matching, with support for SQL wildcard characters.
        starts_with: str
             Parameter to filter the command output based on the string of characters that appear at the beginning of the object name. Uses case-sensitive pattern matching.
        limit: int
             Parameter to limit the maximum number of rows returned by a command.
        from_name: str
             Parameter to enable fetching rows only following the first row whose object name matches the specified string. Case-sensitive and does not have to be the full name.
        deep: bool
             Optionally includes dependency information of the view.
        """
        if show_limit is not None:
            if limit is not None:
                raise ValueError("Cannot specify both 'limit' and 'show_limit'. Use only one of them.")
            warnings.warn(DeprecationWarning("'show_limit' is deprecated, use 'limit' instead."))  # noqa: B028
            limit = show_limit

        resources = self._api.list_views(
            self.database.name,
            self.schema.name,
            like=like,
            starts_with=starts_with,
            show_limit=limit,
            from_name=from_name,
            deep=deep,
            async_req=False,
        )
        return iter(resources)

    @api_telemetry
    def iter_async(
        self,
        *,
        like: Optional[str] = None,
        starts_with: Optional[str] = None,
        limit: Optional[int] = None,
        from_name: Optional[str] = None,
        deep: Optional[bool] = None,
        show_limit: Optional[int] = None,
    ) -> PollingOperation[Iterator[View]]:
        """An asynchronous version of :func:`iter`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """
        if show_limit is not None:
            if limit is not None:
                raise ValueError("Cannot specify both 'limit' and 'show_limit'. Use only one of them.")
            warnings.warn(DeprecationWarning("'show_limit' is deprecated, use 'limit' instead."))  # noqa: B028
            limit = show_limit

        future = self._api.list_views(
            self.database.name,
            self.schema.name,
            like=like,
            starts_with=starts_with,
            show_limit=limit,
            from_name=from_name,
            deep=deep,
            async_req=True,
        )

        return PollingOperations.iterator(future)


class ViewResourceBase(SchemaObjectReferenceMixin["ViewCollection"]):
    _collection_class: ViewCollectionBase
    _identifier_requires_args: bool = False
    _plural_name: str

    def __init__(self, name: str, collection_class: ViewCollectionBase) -> None:
        self.collection = collection_class
        if self._identifier_requires_args:
            self.name_with_args = name
        else:
            self.name = name

    @api_telemetry
    def fetch(
        self,
    ) -> View:
        """Fetch a view.

        Parameters
        __________
        """
        result = self.collection._api.fetch_view(
            self.database.name,
            self.schema.name,
            self._identifier,
            async_req=False,
        )
        return result

    @api_telemetry
    def fetch_async(
        self,
    ) -> PollingOperation[View]:
        """An asynchronous version of :func:`fetch`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """
        future = self.collection._api.fetch_view(
            self.database.name,
            self.schema.name,
            self._identifier,
            async_req=True,
        )

        return PollingOperations.identity(future)

    @api_telemetry
    def get_tags(
        self,
        with_lineage: Optional[bool] = None,
    ) -> dict[TagResource, TagValue]:
        """Get the tag assignments for a view.

        Returns all tags assigned to a view. This operation requires an active warehouse.

        Parameters
        __________
        with_lineage: bool
             Parameter that specifies whether tag assignments inherited by the object from its ancestors in securable object hierarchy should be returned as well: - `true`: All tags assigned to this object should be returned, inheritance included. - `false`: Only tags explicitly assigned to this object should be returned.
        """
        result = self.collection._api.get_tags(
            self.database.name,
            self.schema.name,
            self._identifier,
            with_lineage=with_lineage,
            async_req=False,
        )
        result = dict(self._to_get_tags_tuple(tag_assignment) for tag_assignment in result)
        return result

    @api_telemetry
    def get_tags_async(
        self,
        with_lineage: Optional[bool] = None,
    ) -> PollingOperation[dict[TagResource, TagValue]]:
        """An asynchronous version of :func:`get_tags`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """
        future = self.collection._api.get_tags(
            self.database.name,
            self.schema.name,
            self._identifier,
            with_lineage=with_lineage,
            async_req=True,
        )
        return PollingOperation(
            future, lambda tag_assignments: dict(self._to_get_tags_tuple(ta) for ta in tag_assignments)
        )

    def _to_get_tags_tuple(self, ta: TagAssignment) -> tuple[TagResource, TagValue]:
        db = ta.tag_database
        schema = ta.tag_schema

        if db is None:
            raise ValueError("TagAssignment must have tag_database set.")

        if schema is None:
            raise ValueError("TagAssignment must have tag_schema set.")

        tag_resource = self.root.databases[db].schemas[schema].tags[ta.tag_name]
        tag_value = TagValue(ta.tag_value, ta.level)
        return tag_resource, tag_value

    @api_telemetry
    def set_tags(
        self,
        tags: dict[TagResource, TagValue],
        if_exists: Optional[bool] = None,
    ) -> None:
        """Set tags on a view.

        Parameters
        __________
        tags: dict[TagResource, TagValue]
             (required)
        if_exists: bool
             Parameter that specifies how to handle the request for a resource that does not exist: - `true`: The endpoint does not throw an error if the resource does not exist. It returns a 200 success response, but does not take any action on the resource. - `false`: The endpoint throws an error if the resource doesn't exist.
        """
        tag_assignment = [
            self._to_tag_assignment(tag_resource, tag_value) for [tag_resource, tag_value] in tags.items()
        ]
        self.collection._api.set_tags(
            self.database.name,
            self.schema.name,
            self._identifier,
            tag_assignment=tag_assignment,
            if_exists=if_exists,
            async_req=False,
        )

    @api_telemetry
    def set_tags_async(
        self,
        tags: dict[TagResource, TagValue],
        if_exists: Optional[bool] = None,
    ) -> PollingOperation[None]:
        """An asynchronous version of :func:`set_tags`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """
        tag_assignment = [
            self._to_tag_assignment(tag_resource, tag_value) for [tag_resource, tag_value] in tags.items()
        ]
        future = self.collection._api.set_tags(
            self.database.name,
            self.schema.name,
            self._identifier,
            tag_assignment=tag_assignment,
            if_exists=if_exists,
            async_req=True,
        )
        return PollingOperations.empty(future)

    @staticmethod
    def _to_tag_assignment(tag_resource: TagResource, tag_value: TagValue) -> TagAssignment:
        return TagAssignment(
            tag_name=tag_resource.name,
            tag_schema=tag_resource.schema.name,
            tag_database=tag_resource.database.name,
            tag_value=tag_value.value,
        )

    @api_telemetry
    def unset_tags(
        self,
        tag_resources: set[TagResource],
        if_exists: Optional[bool] = None,
    ) -> None:
        """Unset tags from a view.

        Parameters
        __________
        tag_resources: set[TagResource]
             (required)
        if_exists: bool
             Parameter that specifies how to handle the request for a resource that does not exist: - `true`: The endpoint does not throw an error if the resource does not exist. It returns a 200 success response, but does not take any action on the resource. - `false`: The endpoint throws an error if the resource doesn't exist.
        """
        tag_reference = [self._to_tag_reference(tag_resource) for tag_resource in tag_resources]
        self.collection._api.unset_tags(
            self.database.name,
            self.schema.name,
            self._identifier,
            tag_reference=tag_reference,
            if_exists=if_exists,
            async_req=False,
        )

    @api_telemetry
    def unset_tags_async(
        self,
        tag_resources: set[TagResource],
        if_exists: Optional[bool] = None,
    ) -> PollingOperation[None]:
        """An asynchronous version of :func:`unset_tags`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """
        tag_reference = [self._to_tag_reference(tag_resource) for tag_resource in tag_resources]
        future = self.collection._api.unset_tags(
            self.database.name,
            self.schema.name,
            self._identifier,
            tag_reference=tag_reference,
            if_exists=if_exists,
            async_req=True,
        )
        return PollingOperations.empty(future)

    @staticmethod
    def _to_tag_reference(resource: TagResource) -> TagReference:
        return TagReference(
            tag_name=resource.name,
            tag_schema=resource.schema.name,
            tag_database=resource.database.name,
        )

    @api_telemetry
    def drop(
        self,
        if_exists: Optional[bool] = None,
    ) -> None:
        """Delete a view.

        Parameters
        __________
        if_exists: bool
             Parameter that specifies how to handle the request for a resource that does not exist: - `true`: The endpoint does not throw an error if the resource does not exist. It returns a 200 success response, but does not take any action on the resource. - `false`: The endpoint throws an error if the resource doesn't exist.
        """
        self.collection._api.delete_view(
            self.database.name,
            self.schema.name,
            self._identifier,
            if_exists=if_exists,
            async_req=False,
        )

    @api_telemetry
    def drop_async(
        self,
        if_exists: Optional[bool] = None,
    ) -> PollingOperation[None]:
        """An asynchronous version of :func:`drop`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """
        future = self.collection._api.delete_view(
            self.database.name,
            self.schema.name,
            self._identifier,
            if_exists=if_exists,
            async_req=True,
        )
        return PollingOperations.empty(future)

    @property
    def _identifier(self) -> str:
        """Get the identifier for this resource."""
        if self._identifier_requires_args:
            return self.name_with_args
        return self.name

    def _set_new_name(self, target_name: str) -> None:
        """Set a new name for the resource."""
        if self._identifier_requires_args:
            old_name = self.name_with_args
            self.name_with_args = replace_function_name_in_name_with_args(self.name_with_args, target_name)
        else:
            old_name = self.name
            self.name = target_name

        # Update the collection reference to point to the new name
        self.collection.update_reference(old_name=old_name, new_name=self._identifier, resource=self)



import os

import pytest

from snowflake.core.exceptions import ServerError


pytestmark = [pytest.mark.min_sf_ver("8.35.0")]

CURRENT_DIR = os.path.dirname(os.path.abspath(__file__))


def test_refresh_pipe(pipes, template_pipe, temp_stage):
    try:
        temp_stage.upload_file(CURRENT_DIR + "/../../resources/schema.yaml", "/dir1", auto_compress=False)
        temp_stage.upload_file(CURRENT_DIR + "/../../resources/schema.yaml", "/dir1/dir2", auto_compress=False)
        pipe_resource = pipes.create(template_pipe)

        # refresh pipe in valid format
        pipe_resource.refresh(prefix="dir1/")

        # refresh pipe with non valid dir
        with pytest.raises(ServerError):
            pipe_resource.refresh(prefix="dir3/")
    finally:
        pipe_resource.drop(if_exists=True)



# coding: utf-8

# flake8: noqa

"""
Snowflake Artifact Repository API.

The Snowflake Artifact Repository API is a REST API that you can use to access, update, and perform certain actions on Artifact Repository resource in a Snowflake database.  # noqa: E501

The version of the OpenAPI document: 0.0.1
Contact: support@snowflake.com
Generated by: https://openapi-generator.tech

Do not edit this file manually.
"""

from __future__ import absolute_import

__version__ = "1.0.0"

# import apis into sdk package
from snowflake.core.artifact_repository._generated.api.artifact_repository_api import ArtifactRepositoryApi

# import models into sdk package
from snowflake.core.artifact_repository._generated.models.artifact_repository import ArtifactRepository
from snowflake.core.artifact_repository._generated.models.error_response import ErrorResponse
from snowflake.core.artifact_repository._generated.models.success_accepted_response import SuccessAcceptedResponse
from snowflake.core.artifact_repository._generated.models.success_response import SuccessResponse



from __future__ import absolute_import

# flake8: noqa

# import apis into api package
from snowflake.core.iceberg_table._generated.api.iceberg_table_api import IcebergTableApi

__all__ = [
    "IcebergTableApi",
]



"""Manages Snowflake Functions."""

from ..function._generated.models import Function, FunctionArgument, ServiceFunction
from ._function import FunctionCollection, FunctionResource


__all__ = ["Function", "FunctionArgument", "FunctionCollection", "FunctionResource", "ServiceFunction"]



"""
Cortex Agent API.

OpenAPI 3.0 specification for the Cortex Agent API  # noqa: E501

The version of the OpenAPI document: 0.1.0
Contact: support@snowflake.com
Generated by: https://openapi-generator.tech

Do not edit this file manually.
"""

from __future__ import annotations

import json
import pprint
import re  # noqa: F401

from typing import Any, Dict, List, Optional

from pydantic import BaseModel, ConfigDict, Field, StrictInt, StrictStr, field_validator
from typing_extensions import Annotated

from snowflake.core.cortex.lite_agent_service._generated.models.agent_instructions import (
    AgentInstructions,
    AgentInstructionsModel,
)
from snowflake.core.cortex.lite_agent_service._generated.models.agent_models import AgentModels, AgentModelsModel
from snowflake.core.cortex.lite_agent_service._generated.models.agent_orchestration import (
    AgentOrchestration,
    AgentOrchestrationModel,
)
from snowflake.core.cortex.lite_agent_service._generated.models.mcp_server_wrapper import (
    McpServerWrapper,
    McpServerWrapperModel,
)
from snowflake.core.cortex.lite_agent_service._generated.models.message import Message, MessageModel
from snowflake.core.cortex.lite_agent_service._generated.models.tool import Tool, ToolModel
from snowflake.core.cortex.lite_agent_service._generated.models.tool_choice import ToolChoice, ToolChoiceModel


class AgentRunRequest(BaseModel):
    """A model object representing the AgentRunRequest resource.

    Constructs an object of type AgentRunRequest with the provided properties.

    Parameters
    __________
    messages : list[Message]
        The conversation history and current message.
        Contains both user queries and assistant responses in chronological order.
    models : AgentModels, optional

    orchestration : AgentOrchestration, optional

    instructions : AgentInstructions, optional

    model : str, optional
        The identifier of the LLM to use for processing.
    response_instruction : str, optional
        Optional instructions to guide the model's response style and behavior.
        Can be used to set the tone, format, or specific requirements for responses.
    experimental : object, optional
        reserved
    tools : list[Tool], optional
        List of tools available for the agent to use.
        Tools may have a corresponding configuration in tool_resources.
    tool_resources : Dict[str, object], optional
        Configuration for each tool referenced in the tools array.
        Keys must match the name field of tools.
    tool_choice : ToolChoice, optional

    thread_id : int, optional
        The id of the thread.
    parent_message_id : int, optional
        The id of the message from which this run should begin.
    variables : object, optional
        Dictionary of variables with metadata.
        Each variable can include value, type information, and flags indicating whether it's a session or prompt variable.
    query_tags : list[object], optional
        Query tags to be attached to the Snowflake session. These will be applied for all queries executed by the agent.
    tool_bindings : object, optional
        Maps tools to their variable bindings. For each tool, specifies which input parameters
        should be automatically injected with variable definitions.
    origin_application : str,  default 'external'
        The application that initiated this agent run request.
        Used for tracking and analytics.
    internal_metadata : object, optional
        Internal metadata for the request. This field is reserved for internal use.
    client_metadata : object, optional
        Client-provided metadata for the request. Can be used by clients
        to pass additional context or configuration.
    mcp_servers : list[McpServerWrapper], optional
        List of MCP (Model Context Protocol) servers to use for this request.
        Each server is specified with a server_spec containing its fully qualified name.
    """

    models: Optional[AgentModels] = None

    orchestration: Optional[AgentOrchestration] = None

    instructions: Optional[AgentInstructions] = None

    model: Optional[StrictStr] = None

    response_instruction: Optional[StrictStr] = None

    experimental: Optional[Dict[str, Any]] = None

    messages: Annotated[List[Message], Field(min_length=1)]

    tools: Optional[List[Tool]] = None

    tool_resources: Optional[Dict[str, Dict[str, Any]]] = None

    tool_choice: Optional[ToolChoice] = None

    thread_id: Optional[StrictInt] = None

    parent_message_id: Optional[StrictInt] = None

    variables: Optional[Dict[str, Any]] = None

    query_tags: Optional[List[Any]] = None

    tool_bindings: Optional[Dict[str, Any]] = None

    origin_application: Optional[StrictStr] = "external"

    internal_metadata: Optional[Dict[str, Any]] = None

    client_metadata: Optional[Dict[str, Any]] = None

    mcp_servers: Optional[List[McpServerWrapper]] = None

    __properties = [
        "models",
        "orchestration",
        "instructions",
        "model",
        "response_instruction",
        "experimental",
        "messages",
        "tools",
        "tool_resources",
        "tool_choice",
        "thread_id",
        "parent_message_id",
        "variables",
        "query_tags",
        "tool_bindings",
        "origin_application",
        "internal_metadata",
        "client_metadata",
        "mcp_servers",
    ]

    @field_validator("origin_application")
    def origin_application_validate_enum(cls, v):
        if v is None:
            return v
        if v not in ("inline_copilot", "data_science_agent", "microsoft_teams", "coding_agent", "external"):
            raise ValueError(
                "must validate the enum values ('inline_copilot','data_science_agent','microsoft_teams','coding_agent','external')"
            )
        return v

    model_config = ConfigDict(
        validate_by_name=True,
        validate_assignment=True,
    )

    def to_str(self) -> str:
        """Returns the string representation of the model using alias."""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias."""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> AgentRunRequest:
        """Create an instance of AgentRunRequest from a JSON string."""
        return cls.from_dict(json.loads(json_str))

    def to_dict(
        self,
        hide_readonly_properties: bool = False,
    ) -> dict[str, Any]:
        """Returns the dictionary representation of the model using alias."""
        exclude_properties = set()

        if hide_readonly_properties:
            exclude_properties.update({})

        _dict = self.model_dump(serialize_as_any=True, by_alias=True, exclude=exclude_properties, exclude_none=True)

        # override the default output from pydantic by calling `to_dict()` of models
        if self.models:
            _dict["models"] = self.models.to_dict()

        # override the default output from pydantic by calling `to_dict()` of orchestration
        if self.orchestration:
            _dict["orchestration"] = self.orchestration.to_dict()

        # override the default output from pydantic by calling `to_dict()` of instructions
        if self.instructions:
            _dict["instructions"] = self.instructions.to_dict()

        # override the default output from pydantic by calling `to_dict()` of each item in messages (list)
        _items = []
        if self.messages:
            for _item in self.messages:
                if _item:
                    _items.append(_item.to_dict())
            _dict["messages"] = _items

        # override the default output from pydantic by calling `to_dict()` of each item in tools (list)
        _items = []
        if self.tools:
            for _item in self.tools:
                if _item:
                    _items.append(_item.to_dict())
            _dict["tools"] = _items

        # override the default output from pydantic by calling `to_dict()` of tool_choice
        if self.tool_choice:
            _dict["tool_choice"] = self.tool_choice.to_dict()

        # override the default output from pydantic by calling `to_dict()` of each item in mcp_servers (list)
        _items = []
        if self.mcp_servers:
            for _item in self.mcp_servers:
                if _item:
                    _items.append(_item.to_dict())
            _dict["mcp_servers"] = _items

        # set to None if tool_choice (nullable) is None
        if self.tool_choice is None:
            _dict["tool_choice"] = None

        return _dict

    def to_dict_without_readonly_properties(self) -> dict[str, Any]:
        """Return the dictionary representation of the model without readonly properties."""
        return self.to_dict(hide_readonly_properties=True)

    @classmethod
    def from_dict(cls, obj: dict) -> AgentRunRequest:
        """Create an instance of AgentRunRequest from a dict."""
        if obj is None:
            return None

        if type(obj) is not dict:
            return AgentRunRequest.model_validate(obj)

        _obj = AgentRunRequest.model_validate(
            {
                "models": AgentModels.from_dict(obj.get("models")) if obj.get("models") is not None else None,
                "orchestration": AgentOrchestration.from_dict(obj.get("orchestration"))
                if obj.get("orchestration") is not None
                else None,
                "instructions": AgentInstructions.from_dict(obj.get("instructions"))
                if obj.get("instructions") is not None
                else None,
                "model": obj.get("model"),
                "response_instruction": obj.get("response_instruction"),
                "experimental": obj.get("experimental"),
                "messages": [Message.from_dict(_item) for _item in obj.get("messages")]
                if obj.get("messages") is not None
                else None,
                "tools": [Tool.from_dict(_item) for _item in obj.get("tools")]
                if obj.get("tools") is not None
                else None,
                "tool_resources": obj.get("tool_resources"),
                "tool_choice": ToolChoice.from_dict(obj.get("tool_choice"))
                if obj.get("tool_choice") is not None
                else None,
                "thread_id": obj.get("thread_id"),
                "parent_message_id": obj.get("parent_message_id"),
                "variables": obj.get("variables"),
                "query_tags": obj.get("query_tags"),
                "tool_bindings": obj.get("tool_bindings"),
                "origin_application": obj.get("origin_application")
                if obj.get("origin_application") is not None
                else "external",
                "internal_metadata": obj.get("internal_metadata"),
                "client_metadata": obj.get("client_metadata"),
                "mcp_servers": [McpServerWrapper.from_dict(_item) for _item in obj.get("mcp_servers")]
                if obj.get("mcp_servers") is not None
                else None,
            }
        )

        return _obj


class AgentRunRequestModel:
    def __init__(
        self,
        messages: list[Message],
        # optional properties
        models: Optional[AgentModels] = None,
        orchestration: Optional[AgentOrchestration] = None,
        instructions: Optional[AgentInstructions] = None,
        model: Optional[str] = None,
        response_instruction: Optional[str] = None,
        experimental: Optional[object] = None,
        tools: Optional[list[Tool]] = None,
        tool_resources: Optional[Dict[str, object]] = None,
        tool_choice: Optional[ToolChoice] = None,
        thread_id: Optional[int] = None,
        parent_message_id: Optional[int] = None,
        variables: Optional[object] = None,
        query_tags: Optional[list[object]] = None,
        tool_bindings: Optional[object] = None,
        origin_application: Optional[str] = "external",
        internal_metadata: Optional[object] = None,
        client_metadata: Optional[object] = None,
        mcp_servers: Optional[list[McpServerWrapper]] = None,
    ):
        """A model object representing the AgentRunRequest resource.

                Constructs an object of type AgentRunRequest with the provided properties.

                Parameters
                __________
                messages : list[Message]
                    The conversation history and current message.
        Contains both user queries and assistant responses in chronological order.

                models : AgentModels, optional

                orchestration : AgentOrchestration, optional

                instructions : AgentInstructions, optional

                model : str, optional
                    The identifier of the LLM to use for processing.
                response_instruction : str, optional
                    Optional instructions to guide the model's response style and behavior.
        Can be used to set the tone, format, or specific requirements for responses.

                experimental : object, optional
                    reserved
                tools : list[Tool], optional
                    List of tools available for the agent to use.
        Tools may have a corresponding configuration in tool_resources.

                tool_resources : Dict[str, object], optional
                    Configuration for each tool referenced in the tools array.
        Keys must match the name field of tools.

                tool_choice : ToolChoice, optional

                thread_id : int, optional
                    The id of the thread.

                parent_message_id : int, optional
                    The id of the message from which this run should begin.

                variables : object, optional
                    Dictionary of variables with metadata.
        Each variable can include value, type information, and flags indicating whether it's a session or prompt variable.

                query_tags : list[object], optional
                    Query tags to be attached to the Snowflake session. These will be applied for all queries executed by the agent.

                tool_bindings : object, optional
                    Maps tools to their variable bindings. For each tool, specifies which input parameters
        should be automatically injected with variable definitions.

                origin_application : str,  default 'external'
                    The application that initiated this agent run request.
        Used for tracking and analytics.

                internal_metadata : object, optional
                    Internal metadata for the request. This field is reserved for internal use.

                client_metadata : object, optional
                    Client-provided metadata for the request. Can be used by clients
        to pass additional context or configuration.

                mcp_servers : list[McpServerWrapper], optional
                    List of MCP (Model Context Protocol) servers to use for this request.
        Each server is specified with a server_spec containing its fully qualified name.
        """
        self.models = models
        self.orchestration = orchestration
        self.instructions = instructions
        self.model = model
        self.response_instruction = response_instruction
        self.experimental = experimental
        self.messages = messages
        self.tools = tools
        self.tool_resources = tool_resources
        self.tool_choice = tool_choice
        self.thread_id = thread_id
        self.parent_message_id = parent_message_id
        self.variables = variables
        self.query_tags = query_tags
        self.tool_bindings = tool_bindings
        self.origin_application = origin_application
        self.internal_metadata = internal_metadata
        self.client_metadata = client_metadata
        self.mcp_servers = mcp_servers

    __properties = [
        "models",
        "orchestration",
        "instructions",
        "model",
        "response_instruction",
        "experimental",
        "messages",
        "tools",
        "tool_resources",
        "tool_choice",
        "thread_id",
        "parent_message_id",
        "variables",
        "query_tags",
        "tool_bindings",
        "origin_application",
        "internal_metadata",
        "client_metadata",
        "mcp_servers",
    ]

    def __repr__(self) -> str:
        return repr(self._to_model())

    def _to_model(self):
        return AgentRunRequest(
            models=self.models._to_model() if self.models is not None else None,
            orchestration=self.orchestration._to_model() if self.orchestration is not None else None,
            instructions=self.instructions._to_model() if self.instructions is not None else None,
            model=self.model,
            response_instruction=self.response_instruction,
            experimental=self.experimental,
            messages=[x._to_model() for x in self.messages] if self.messages is not None else None,
            tools=[x._to_model() for x in self.tools] if self.tools is not None else None,
            tool_resources=self.tool_resources,
            tool_choice=self.tool_choice._to_model() if self.tool_choice is not None else None,
            thread_id=self.thread_id,
            parent_message_id=self.parent_message_id,
            variables=self.variables,
            query_tags=self.query_tags,
            tool_bindings=self.tool_bindings,
            origin_application=self.origin_application,
            internal_metadata=self.internal_metadata,
            client_metadata=self.client_metadata,
            mcp_servers=[x._to_model() for x in self.mcp_servers] if self.mcp_servers is not None else None,
        )

    @classmethod
    def _from_model(cls, model) -> AgentRunRequestModel:
        return AgentRunRequestModel(
            models=AgentModelsModel._from_model(model.models) if model.models else None,
            orchestration=AgentOrchestrationModel._from_model(model.orchestration) if model.orchestration else None,
            instructions=AgentInstructionsModel._from_model(model.instructions) if model.instructions else None,
            model=model.model,
            response_instruction=model.response_instruction,
            experimental=model.experimental,
            messages=[MessageModel._from_model(x) for x in model.messages] if model.messages else None,
            tools=[ToolModel._from_model(x) for x in model.tools] if model.tools else None,
            tool_resources=model.tool_resources,
            tool_choice=ToolChoiceModel._from_model(model.tool_choice) if model.tool_choice else None,
            thread_id=model.thread_id,
            parent_message_id=model.parent_message_id,
            variables=model.variables,
            query_tags=model.query_tags,
            tool_bindings=model.tool_bindings,
            origin_application=model.origin_application,
            internal_metadata=model.internal_metadata,
            client_metadata=model.client_metadata,
            mcp_servers=[McpServerWrapperModel._from_model(x) for x in model.mcp_servers]
            if model.mcp_servers
            else None,
        )

    def to_dict(self):
        """Create a dictionary of the properties from a AgentRunRequest.

        This method constructs a dictionary with the key-value entries corresponding to the properties of the AgentRunRequest object.

        Returns
        _______
        dict
            A dictionary object created using the input model.
        """
        return self._to_model().to_dict()

    @classmethod
    def from_dict(cls, obj: dict) -> AgentRunRequestModel:
        """Create an instance of AgentRunRequest from a dict.

        This method constructs a AgentRunRequest object from a dictionary with the key-value pairs of its properties.

        Parameters
        ----------
        obj : dict
            A dictionary whose keys and values correspond to the properties of the resource object.

        Returns
        _______
        AgentRunRequest
            A AgentRunRequest object created using the input dictionary; this will fail if the required properties are missing.
        """
        return cls._from_model(AgentRunRequest.from_dict(obj))


AgentRunRequest._model_class = AgentRunRequestModel



import copy

from contextlib import suppress

import pytest as pytest

from snowflake.core.exceptions import NotFoundError
from tests.integ.utils import random_string

from .conftest import test_view_template


@pytest.mark.min_sf_ver("8.33.0")
def test_create_and_fetch_view(views, temp_table):
    view_name = random_string(10, "test_view_")
    view_handle = views[view_name]
    try:
        test_view = copy.deepcopy(test_view_template)
        test_view.name = view_name
        test_view.query = f"SELECT * FROM {temp_table.name}"
        views.create(test_view)

        fetch_handle = view_handle.fetch()
        assert fetch_handle.name.upper() == view_name.upper()
        assert len(fetch_handle.columns) == 3
        assert fetch_handle.columns[0].name == "C1"
        assert fetch_handle.columns[0].datatype.upper() == "NUMBER(38,0)"
        assert fetch_handle.columns[1].name == "C2"
        assert fetch_handle.columns[1].datatype.upper() == "VARCHAR(16777216)"
        assert fetch_handle.columns[2].name == "C3"
        assert fetch_handle.columns[2].datatype.upper() == "VARCHAR(16777216)"

    finally:
        with suppress(NotFoundError):
            view_handle.drop()



"""
Cortex Embed API.

OpenAPI 3.0 specification for the Cortex Embed REST API  # noqa: E501

The version of the OpenAPI document: 0.1.0
Contact: support@snowflake.com
Generated by: https://openapi-generator.tech

Do not edit this file manually.
"""

import json
import logging
import os
import re
import tempfile

from concurrent.futures import Future
from datetime import date, datetime
from typing import Literal, Union, overload

from dateutil.parser import parse
from pydantic import Field, validate_call
from typing_extensions import Annotated

import snowflake.core._http_requests

from snowflake.core._generated import rest
from snowflake.core._generated.api_client import ApiClient
from snowflake.core._internal.utils import is_running_inside_stored_procedure
from snowflake.core.cortex.embed_service._generated import models
from snowflake.core.cortex.embed_service._generated.models.embed_request import EmbedRequest
from snowflake.core.cortex.embed_service._generated.models.embed_response import EmbedResponse
from snowflake.core.exceptions import (  # noqa: F401
    _APITypeError,
    _APIValueError,
)


logger = logging.getLogger(__name__)


class CortexEmbedApi:
    """NOTE: This class is auto generated by OpenAPI Generator.

    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    def __init__(self, root, resource_class, sproc_client):
        self._root = root
        self._resource_name = "embed_service"
        self._resource_class = resource_class

        if is_running_inside_stored_procedure():
            self._api_client = sproc_client
        else:
            self._api_client = ApiClient.get_default()

    @property
    def api_client(self):
        return self._api_client

    def deserialize(self, response, response_type):
        """Deserializes response into an object.

        :param response: RESTResponse object to be deserialized.
        :param response_type: class literal for
            deserialized object, or string of class name.

        :return: deserialized object.
        """
        # handle file downloading
        # save response body into a tmp file and return the instance
        if response_type == "file":
            return self.__deserialize_file(response)

        # fetch data from response object
        try:
            data = json.loads(response.data)
        except ValueError:
            data = response.data

        return self.__deserialize(data, response_type)

    def __deserialize(self, data, klass):
        """Deserializes dict, list, str into an object.

        :param data: dict, list or str.
        :param klass: class literal, or string of class name.

        :return: object.
        """
        if data is None:
            return None

        if isinstance(klass, str):
            if klass.startswith("Iterable["):
                sub_kls = re.match(r"Iterable\[(.*)]", klass).group(1)
                return [self.__deserialize(sub_data, sub_kls) for sub_data in data]

            if klass.startswith("Dict["):
                sub_kls = re.match(r"Dict\[([^,]*), (.*)]", klass).group(2)
                return {k: self.__deserialize(v, sub_kls) for k, v in data.items()}

            # convert str to class
            if klass in snowflake.core._http_requests.NATIVE_TYPES_MAPPING:
                klass = snowflake.core._http_requests.NATIVE_TYPES_MAPPING[klass]
            else:
                klass = getattr(models, klass)

        if klass in snowflake.core._http_requests.PRIMITIVE_TYPES:
            return self.__deserialize_primitive(data, klass)
        elif klass == object:  # noqa: E721
            return self.__deserialize_object(data)
        elif klass == date:
            return self.__deserialize_date(data)
        elif klass == datetime:
            return self.__deserialize_datetime(data)
        else:
            return self.__deserialize_model(data, klass)

    def __deserialize_file(self, response):
        """Deserializes body to file.

        Saves response body into a file in a temporary folder,
        using the filename from the `Content-Disposition` header if provided.

        :param response:  RESTResponse.
        :return: file path.
        """
        fd, path = tempfile.mkstemp(dir=self.api_client.configuration.temp_folder_path)
        os.close(fd)
        os.remove(path)

        content_disposition = response.getheader("Content-Disposition")
        if content_disposition:
            filename = re.search(r'filename=[\'"]?([^\'"\s]+)[\'"]?', content_disposition).group(1)
            path = os.path.join(os.path.dirname(path), filename)

        with open(path, "wb") as f:
            f.write(response.data)

        return path

    def __deserialize_primitive(self, data, klass):
        """Deserializes string to primitive type.

        :param data: str.
        :param klass: class literal.

        :return: int, long, float, str, bool.
        """
        try:
            return klass(data)
        except UnicodeEncodeError:
            return str(data)
        except TypeError:
            return data

    def __deserialize_object(self, value):
        """Return an original value.

        :return: object.
        """
        return value

    def __deserialize_date(self, string):
        """Deserializes string to date.

        :param string: str.
        :return: date.
        """
        try:
            return parse(string).date()
        except ImportError:
            return string
        except ValueError:
            raise rest.APIError(status=0, reason=f"Failed to parse `{string}` as date object")

    def __deserialize_datetime(self, string):
        """Deserializes string to datetime.

        The string should be in iso8601 datetime format.

        :param string: str.
        :return: datetime.
        """
        try:
            return parse(string)
        except ImportError:
            return string
        except ValueError:
            raise rest.APIError(status=0, reason=(f"Failed to parse `{string}` as datetime object"))

    def __deserialize_model(self, data, klass):
        """Deserializes list or dict to model.

        :param data: dict, list.
        :param klass: class literal.
        :return: model object.
        """
        return klass.from_dict(data)

    @overload
    def embed(
        self,
        embed_request: Annotated[
            EmbedRequest, Field(description="Request payload containing model identifier and input text.")
        ],
        async_req: Literal[False] = False,
        **kwargs,
    ) -> EmbedResponse: ...

    @overload
    def embed(
        self,
        embed_request: Annotated[
            EmbedRequest, Field(description="Request payload containing model identifier and input text.")
        ],
        async_req: Literal[True] = True,
        **kwargs,
    ) -> Future[EmbedResponse]: ...

    @overload
    def embed(
        self,
        embed_request: Annotated[
            EmbedRequest, Field(description="Request payload containing model identifier and input text.")
        ],
        async_req: bool = False,
        **kwargs,
    ) -> Union[EmbedResponse, Future[EmbedResponse]]: ...

    @validate_call
    def embed(
        self,
        embed_request: Annotated[
            EmbedRequest, Field(description="Request payload containing model identifier and input text.")
        ],
        **kwargs,
    ) -> Union[EmbedResponse, Future[EmbedResponse]]:
        r"""Perform LLM embed  # noqa: E501

        Perform LLM embedding for input text, similar to the Snowflake Cortex `EMBED_TEXT` functions. For more details on available functions, see: - [EMBED_TEXT_768](https://docs.snowflake.com/en/sql-reference/functions/embed_text-snowflake-cortex) - [EMBED_TEXT_1024](https://docs.snowflake.com/en/sql-reference/functions/embed_text_1024-snowflake-cortex)   # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> future = api.embed(embed_request, async_req=True)
        >>> result = future.result()
        :param embed_request: Request payload containing model identifier and input text. (required)
        :type embed_request: EmbedRequest
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _preload_content: if False, the urllib3.HTTPResponse object will
                                 be returned without reading/decoding response
                                 data. Default is True.
        :type _preload_content: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns a Future object representing the execution of the method.
        :rtype: Union[EmbedResponse, Future[EmbedResponse]]
        """
        kwargs["_return_http_data_only"] = True
        return self.embed_with_http_info(embed_request, **kwargs)

    @validate_call
    def embed_with_http_info(
        self,
        embed_request: Annotated[
            EmbedRequest, Field(description="Request payload containing model identifier and input text.")
        ],
        **kwargs,
    ):
        r"""Perform LLM embed  # noqa: E501

        Perform LLM embedding for input text, similar to the Snowflake Cortex `EMBED_TEXT` functions. For more details on available functions, see: - [EMBED_TEXT_768](https://docs.snowflake.com/en/sql-reference/functions/embed_text-snowflake-cortex) - [EMBED_TEXT_1024](https://docs.snowflake.com/en/sql-reference/functions/embed_text_1024-snowflake-cortex)   # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> future = api.embed_with_http_info(embed_request, async_req=True)
        >>> result = future.result()
        :param embed_request: Request payload containing model identifier and input text. (required)
        :type embed_request: EmbedRequest
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _return_http_data_only: response data without head status code
                                       and headers
        :type _return_http_data_only: bool, optional
        :param _preload_content: if False, the urllib3.HTTPResponse object will
                                 be returned without reading/decoding response
                                 data. Default is True.
        :type _preload_content: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the authentication
                              in the spec for a single request.
        :type _request_auth: dict, optional
        :type _content_type: string, optional: force content-type for the request
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns a Future object representing the execution of the method.
        :rtype: tuple(Union[EmbedResponse, Future[EmbedResponse]], status_code(int), headers(HTTPHeaderDict))
        """
        _params = locals()

        _all_params = ["embed_request"]
        _all_params.extend(
            [
                "async_req",
                "_return_http_data_only",
                "_preload_content",
                "_request_timeout",
                "_request_auth",
                "_content_type",
                "_headers",
            ]
        )

        # validate the arguments
        for _key, _val in _params["kwargs"].items():
            if _key not in _all_params:
                raise _APITypeError(f"Got an unexpected keyword argument '{_key}' to method embed")
            _params[_key] = _val
        del _params["kwargs"]

        _collection_formats = {}

        # process the path parameters
        _path_params = {}

        # process the query parameters
        _query_params = []

        # process the header parameters
        _header_params = dict(_params.get("_headers", {}))

        # process the form parameters
        _form_params = []
        _files = {}

        # process the body parameter
        _body_params = None

        if _params["embed_request"] is not None:
            _body_params = _params["embed_request"]

        # set the HTTP header `Accept`
        _header_params["Accept"] = self.api_client.select_header_accept(["application/json"])

        # set the HTTP header `Content-Type`
        _content_types_list = _params.get(
            "_content_type", self.api_client.select_header_content_type(["application/json"])
        )
        if _content_types_list:
            _header_params["Content-Type"] = _content_types_list

        # authentication setting
        _auth_settings = ["ExternalOAuth", "KeyPair", "ProgrammaticAccessToken", "SnowflakeOAuth"]

        _response_types_map = {
            "200": "EmbedResponse",
            "400": "ErrorResponse",
            "401": "ErrorResponse",
            "403": "ErrorResponse",
            "404": "ErrorResponse",
            "405": "ErrorResponse",
            "500": "ErrorResponse",
            "503": "ErrorResponse",
            "504": "ErrorResponse",
        }

        return self.api_client.call_api(
            self._root,
            "/api/v2/cortex/inference:embed",
            "POST",
            _path_params,
            _query_params,
            _header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            response_types_map=_response_types_map,
            auth_settings=_auth_settings,
            async_req=_params.get("async_req"),
            _return_http_data_only=_params.get("_return_http_data_only"),
            _preload_content=_params.get("_preload_content", True),
            _request_timeout=_params.get("_request_timeout"),
            collection_formats=_collection_formats,
            _request_auth=_params.get("_request_auth"),
            _deserialize_response_fn=self.deserialize,
        )



import logging
import types

from threading import Event
from unittest import mock
from unittest.mock import MagicMock, patch

import pytest

from snowflake.core._internal.snowapi_parameters import SnowApiParameter, SnowApiParameters
from snowflake.core._root import Root
from snowflake.core.database import DatabaseCollection
from snowflake.core.table._generated.models.table import Table
from snowflake.core.table._generated.models.table_column import TableColumn


@pytest.fixture
def fake_root():
    """Mock for Root.

    Usage of this central definition is necessary since the underlying
    generated Configuration class is handled as a singleton, so we treat
    the unit test root as a singleton as well.
    """
    with patch("snowflake.core._root.Root"):
        mock_instance = MagicMock()
        mock_instance._hostname = "localhost"
        mock_instance.connection = mock.MagicMock(rest=mock.MagicMock(_protocol="http", _port="80"))
        mock_instance.root_config = MagicMock()
        mock_instance.root_config.has_user_agents = MagicMock(return_value=False)
        mock_instance.root_config.get_user_agents = MagicMock(return_value="")
        mock_instance.parameters.return_value = SnowApiParameters({SnowApiParameter.MAX_THREADS: "1"})

        mock_instance._initialize_client_info = types.MethodType(Root._initialize_client_info, mock_instance)

        return mock_instance


@pytest.fixture()
def logger_level_info(caplog):
    # Default logger level to info

    with caplog.at_level(logging.INFO):
        yield


@pytest.fixture
def dbs(fake_root):
    return DatabaseCollection(fake_root)


@pytest.fixture
def db(dbs):
    return dbs["my_db"]


@pytest.fixture
def schemas(db):
    return db.schemas


@pytest.fixture
def schema(schemas):
    return schemas["my_schema"]


@pytest.fixture
def tables(schema):
    return schema.tables


@pytest.fixture
def stages(schema):
    return schema.stages


@pytest.fixture
def table(tables):
    return Table(name="my_table", columns=[TableColumn(name="c1", datatype="int")])


@pytest.fixture()
def event():
    event = Event()
    yield event
    event.set()


@pytest.fixture
def tags(schema):
    return schema.tags


@pytest.fixture
def tag(tags):
    return tags["my_tag"]



#
# Copyright (c) 2012-2023 Snowflake Computing Inc. All rights reserved.
#


import pytest

from snowflake.core.exceptions import NotFoundError
from snowflake.core.image_repository import ImageRepository
from tests.utils import random_string


def test_drop(image_repositories):
    try:
        ir_name = random_string(5, "test_ir_")
        test_ir = ImageRepository(name=ir_name)
        image_repositories.create(test_ir)
        image_repositories[test_ir.name].drop()
        with pytest.raises(NotFoundError):
            image_repositories[test_ir.name].fetch()

        # creating again, making sure it's not an issue
        image_repositories.create(test_ir)
        image_repositories[test_ir.name].drop()
    finally:
        image_repositories[test_ir.name].drop(if_exists=True)



"""
Common generated classes shared across all resources.

The version of the OpenAPI document: 0.0.1
Contact: support@snowflake.com
Generated by: https://openapi-generator.tech

Do not edit this file manually.
"""

import logging

from typing import Callable, Generic, Iterable, Iterator, TypeVar, overload


"""Defines an Iterator to represent certain resource instances fetched from the Snowflake database."""


logger = logging.getLogger(__name__)

T = TypeVar("T")
S = TypeVar("S")


class PagedIter(Iterable[T], Generic[T]):
    """A page-by-page iterator.

    Data fetched from the server is iterated over page by page, yielding items one by
    one.

    This iterator works by accepting a closure/partial function, which it will delegate to
    in order to fetch the next page, passing in only the chunk/page index it wants to get.

    Example:
        >>> from snowflake.core import Root
        >>> root = Root(connection)
        >>> tasks: TaskCollection = root.databases["mydb"].schemas["myschema"].tasks
        >>> task_iter = tasks.iter(like="my%")  # returns a PagedIter[Task]
        >>> for task_obj in task_iter:
        ...     print(task_obj.name)
    """

    @overload
    def __init__(self, data: Iterable[T]) -> None: ...

    @overload
    def __init__(self, data: Iterable[T], map_: None) -> None: ...

    @overload
    def __init__(self, data: Iterable[S], map_: Callable[[S], T]) -> None: ...

    def __init__(
        self,
        page_fetch_closure_,
        number_of_chunks_=1,
    ) -> None:
        self._page_fetch_closure = page_fetch_closure_
        self._number_of_chunks = number_of_chunks_
        self._iter = iter(self)
        logger.debug("created a PagedIter for %d chunks", number_of_chunks_)

    def __iter__(self) -> Iterator[T]:
        for chunk in range(self._number_of_chunks):
            logger.debug("started consuming chunk #%d of results", chunk)
            yield from self._page_fetch_closure(chunk_index=chunk)



import pytest

from snowflake.core.api_integration import ApiIntegration, GitHook
from tests.integ.fixtures.constants import UUID
from tests.integ.utils import random_string


@pytest.fixture(scope="module", autouse=True)
def setup(api_integrations):
    names = [random_string(10, f"test_api_integration_iter_a_{UUID}_")]
    for _ in range(2):
        names.append(random_string(10, f"test_api_integration_iter_b_{UUID}_"))
    for _ in range(3):
        names.append(random_string(10, f"test_api_integration_iter_c_{UUID}_"))
    try:
        for name in names:
            ai_def = ApiIntegration(
                name=name,
                api_hook=GitHook(),
                api_allowed_prefixes=["https://github.com"],
                enabled=True,
                comment="created by test_iter",
            )
            api_integrations.create(ai_def)
        yield
    finally:
        for name in names:
            api_integrations[name].drop(if_exists=True)


def test_iter(api_integrations):
    """Test iterating over API integrations, filtering to only our test integrations."""
    test_integrations = list(api_integrations.iter(like="test_api_integration_iter_%"))
    assert len(test_integrations) >= 6


def test_iter_like(api_integrations):
    """Test iterating over API integrations with like filters."""
    # Test non-matching pattern
    x_results = list(api_integrations.iter(like=f"test_api_integration_iter_x_{UUID}_%"))
    assert len(x_results) == 0

    # Test patterns that match our created integrations
    a_results = list(api_integrations.iter(like=f"test_api_integration_iter_a_{UUID}_%"))
    assert len(a_results) == 1

    b_results = list(api_integrations.iter(like=f"test_api_integration_iter_b_{UUID}_%"))
    assert len(b_results) == 2

    c_results = list(api_integrations.iter(like=f"test_api_integration_iter_c_{UUID}_%"))
    assert len(c_results) == 3



"""
Snowflake Password Policy API.

The Snowflake Password Policy API is a REST API that you can use to access, update, and perform certain actions on Password Policy resource in a Snowflake database.  # noqa: E501

The version of the OpenAPI document: 0.0.1
Contact: support@snowflake.com
Generated by: https://openapi-generator.tech

Do not edit this file manually.
"""

from __future__ import annotations

import json
import pprint
import re

from datetime import datetime
from typing import Any, Optional

from pydantic import BaseModel, ConfigDict, Field, StrictInt, StrictStr, field_validator
from typing_extensions import Annotated


class PasswordPolicy(BaseModel):
    """A model object representing the PasswordPolicy resource.

    Constructs an object of type PasswordPolicy with the provided properties.

    Parameters
    __________
    name : str
        Name of the password policy
    comment : str, optional
        Comment for the password policy
    password_min_length : int, optional
        Minimum length of new password.
    password_max_length : int, optional
        Maximum length of new password.
    password_min_upper_case_chars : int, optional
        Minimum number of uppercase characters in new password.
    password_min_lower_case_chars : int, optional
        Minimum number of lowercase characters in new password.
    password_min_numeric_chars : int, optional
        Minimum number of numeric characters in new password.
    password_min_special_chars : int, optional
        Minimum number of special characters in new password.
    password_min_age_days : int, optional
        Period after a password is changed during which a password cannot be changed again, in days.
    password_max_age_days : int, optional
        Period after which password must be changed, in days.
    password_max_retries : int, optional
        Number of attempts users have to enter the correct password before their account is locked.
    password_lockout_time_mins : int, optional
        Period of time for which users will be locked after entering their password incorrectly many times (specified by MAX_RETRIES), in minutes
    password_history : int, optional
        Number of distinct passwords that a user must create before re-using a previous password
    created_on : datetime, optional
        Date and time when the password policy was created — **Read-only:** *any user-provided value will be ignored.*
    database_name : str, optional
        Database in which the password policy is stored — **Read-only:** *any user-provided value will be ignored.*
    schema_name : str, optional
        Schema in which the password policy is stored — **Read-only:** *any user-provided value will be ignored.*
    owner : str, optional
        Role that owns the password policy — **Read-only:** *any user-provided value will be ignored.*
    owner_role_type : str, optional
        The type of role that owns the password policy — **Read-only:** *any user-provided value will be ignored.*
    """

    comment: Optional[StrictStr] = None

    password_min_length: Optional[StrictInt] = None

    password_max_length: Optional[StrictInt] = None

    password_min_upper_case_chars: Optional[StrictInt] = None

    password_min_lower_case_chars: Optional[StrictInt] = None

    password_min_numeric_chars: Optional[StrictInt] = None

    password_min_special_chars: Optional[StrictInt] = None

    password_min_age_days: Optional[StrictInt] = None

    password_max_age_days: Optional[StrictInt] = None

    password_max_retries: Optional[StrictInt] = None

    password_lockout_time_mins: Optional[StrictInt] = None

    password_history: Optional[StrictInt] = None

    name: Annotated[str, Field(strict=True)]

    created_on: Optional[datetime] = None

    database_name: Optional[Annotated[str, Field(strict=True)]] = None

    schema_name: Optional[Annotated[str, Field(strict=True)]] = None

    owner: Optional[Annotated[str, Field(strict=True)]] = None

    owner_role_type: Optional[StrictStr] = None

    __properties = [
        "comment",
        "password_min_length",
        "password_max_length",
        "password_min_upper_case_chars",
        "password_min_lower_case_chars",
        "password_min_numeric_chars",
        "password_min_special_chars",
        "password_min_age_days",
        "password_max_age_days",
        "password_max_retries",
        "password_lockout_time_mins",
        "password_history",
        "name",
        "created_on",
        "database_name",
        "schema_name",
        "owner",
        "owner_role_type",
    ]

    @field_validator("name")
    def name_validate_regular_expression(cls, v):
        if not re.match(r"""^\"([^\"]|\"\")+\"|[a-zA-Z_][a-zA-Z0-9_$]*$""", v):
            raise ValueError(r"""must validate the regular expression /^"([^"]|"")+"|[a-zA-Z_][a-zA-Z0-9_$]*$/""")
        return v

    @field_validator("database_name")
    def database_name_validate_regular_expression(cls, v):
        if v is None:
            return v
        if not re.match(r"""^\"([^\"]|\"\")+\"|[a-zA-Z_][a-zA-Z0-9_$]*$""", v):
            raise ValueError(r"""must validate the regular expression /^"([^"]|"")+"|[a-zA-Z_][a-zA-Z0-9_$]*$/""")
        return v

    @field_validator("schema_name")
    def schema_name_validate_regular_expression(cls, v):
        if v is None:
            return v
        if not re.match(r"""^\"([^\"]|\"\")+\"|[a-zA-Z_][a-zA-Z0-9_$]*$""", v):
            raise ValueError(r"""must validate the regular expression /^"([^"]|"")+"|[a-zA-Z_][a-zA-Z0-9_$]*$/""")
        return v

    @field_validator("owner")
    def owner_validate_regular_expression(cls, v):
        if v is None:
            return v
        if not re.match(r"""^\"([^\"]|\"\")+\"|[a-zA-Z_][a-zA-Z0-9_$]*$""", v):
            raise ValueError(r"""must validate the regular expression /^"([^"]|"")+"|[a-zA-Z_][a-zA-Z0-9_$]*$/""")
        return v

    model_config = ConfigDict(
        validate_by_name=True,
        validate_assignment=True,
    )

    def to_str(self) -> str:
        """Returns the string representation of the model using alias."""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias."""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> PasswordPolicy:
        """Create an instance of PasswordPolicy from a JSON string."""
        return cls.from_dict(json.loads(json_str))

    def to_dict(
        self,
        hide_readonly_properties: bool = False,
    ) -> dict[str, Any]:
        """Returns the dictionary representation of the model using alias."""
        exclude_properties = set()

        if hide_readonly_properties:
            exclude_properties.update(
                {
                    "created_on",
                    "database_name",
                    "schema_name",
                    "owner",
                    "owner_role_type",
                }
            )

        _dict = self.model_dump(serialize_as_any=True, by_alias=True, exclude=exclude_properties, exclude_none=True)

        return _dict

    def to_dict_without_readonly_properties(self) -> dict[str, Any]:
        """Return the dictionary representation of the model without readonly properties."""
        return self.to_dict(hide_readonly_properties=True)

    @classmethod
    def from_dict(cls, obj: dict) -> PasswordPolicy:
        """Create an instance of PasswordPolicy from a dict."""
        if obj is None:
            return None

        if type(obj) is not dict:
            return PasswordPolicy.model_validate(obj)

        _obj = PasswordPolicy.model_validate(
            {
                "comment": obj.get("comment"),
                "password_min_length": obj.get("password_min_length"),
                "password_max_length": obj.get("password_max_length"),
                "password_min_upper_case_chars": obj.get("password_min_upper_case_chars"),
                "password_min_lower_case_chars": obj.get("password_min_lower_case_chars"),
                "password_min_numeric_chars": obj.get("password_min_numeric_chars"),
                "password_min_special_chars": obj.get("password_min_special_chars"),
                "password_min_age_days": obj.get("password_min_age_days"),
                "password_max_age_days": obj.get("password_max_age_days"),
                "password_max_retries": obj.get("password_max_retries"),
                "password_lockout_time_mins": obj.get("password_lockout_time_mins"),
                "password_history": obj.get("password_history"),
                "name": obj.get("name"),
                "created_on": obj.get("created_on"),
                "database_name": obj.get("database_name"),
                "schema_name": obj.get("schema_name"),
                "owner": obj.get("owner"),
                "owner_role_type": obj.get("owner_role_type"),
            }
        )

        return _obj


class PasswordPolicyModel:
    def __init__(
        self,
        name: str,
        # optional properties
        comment: Optional[str] = None,
        password_min_length: Optional[int] = None,
        password_max_length: Optional[int] = None,
        password_min_upper_case_chars: Optional[int] = None,
        password_min_lower_case_chars: Optional[int] = None,
        password_min_numeric_chars: Optional[int] = None,
        password_min_special_chars: Optional[int] = None,
        password_min_age_days: Optional[int] = None,
        password_max_age_days: Optional[int] = None,
        password_max_retries: Optional[int] = None,
        password_lockout_time_mins: Optional[int] = None,
        password_history: Optional[int] = None,
        created_on: Optional[datetime] = None,
        database_name: Optional[str] = None,
        schema_name: Optional[str] = None,
        owner: Optional[str] = None,
        owner_role_type: Optional[str] = None,
    ):
        """A model object representing the PasswordPolicy resource.

        Constructs an object of type PasswordPolicy with the provided properties.

        Parameters
        __________
        name : str
            Name of the password policy
        comment : str, optional
            Comment for the password policy
        password_min_length : int, optional
            Minimum length of new password.
        password_max_length : int, optional
            Maximum length of new password.
        password_min_upper_case_chars : int, optional
            Minimum number of uppercase characters in new password.
        password_min_lower_case_chars : int, optional
            Minimum number of lowercase characters in new password.
        password_min_numeric_chars : int, optional
            Minimum number of numeric characters in new password.
        password_min_special_chars : int, optional
            Minimum number of special characters in new password.
        password_min_age_days : int, optional
            Period after a password is changed during which a password cannot be changed again, in days.
        password_max_age_days : int, optional
            Period after which password must be changed, in days.
        password_max_retries : int, optional
            Number of attempts users have to enter the correct password before their account is locked.
        password_lockout_time_mins : int, optional
            Period of time for which users will be locked after entering their password incorrectly many times (specified by MAX_RETRIES), in minutes
        password_history : int, optional
            Number of distinct passwords that a user must create before re-using a previous password
        created_on : datetime, optional
            Date and time when the password policy was created.
        database_name : str, optional
            Database in which the password policy is stored
        schema_name : str, optional
            Schema in which the password policy is stored
        owner : str, optional
            Role that owns the password policy
        owner_role_type : str, optional
            The type of role that owns the password policy
        """
        self.comment = comment
        self.password_min_length = password_min_length
        self.password_max_length = password_max_length
        self.password_min_upper_case_chars = password_min_upper_case_chars
        self.password_min_lower_case_chars = password_min_lower_case_chars
        self.password_min_numeric_chars = password_min_numeric_chars
        self.password_min_special_chars = password_min_special_chars
        self.password_min_age_days = password_min_age_days
        self.password_max_age_days = password_max_age_days
        self.password_max_retries = password_max_retries
        self.password_lockout_time_mins = password_lockout_time_mins
        self.password_history = password_history
        self.name = name
        self.created_on = created_on
        self.database_name = database_name
        self.schema_name = schema_name
        self.owner = owner
        self.owner_role_type = owner_role_type

    __properties = [
        "comment",
        "password_min_length",
        "password_max_length",
        "password_min_upper_case_chars",
        "password_min_lower_case_chars",
        "password_min_numeric_chars",
        "password_min_special_chars",
        "password_min_age_days",
        "password_max_age_days",
        "password_max_retries",
        "password_lockout_time_mins",
        "password_history",
        "name",
        "created_on",
        "database_name",
        "schema_name",
        "owner",
        "owner_role_type",
    ]

    def __repr__(self) -> str:
        return repr(self._to_model())

    def _to_model(self):
        return PasswordPolicy(
            comment=self.comment,
            password_min_length=self.password_min_length,
            password_max_length=self.password_max_length,
            password_min_upper_case_chars=self.password_min_upper_case_chars,
            password_min_lower_case_chars=self.password_min_lower_case_chars,
            password_min_numeric_chars=self.password_min_numeric_chars,
            password_min_special_chars=self.password_min_special_chars,
            password_min_age_days=self.password_min_age_days,
            password_max_age_days=self.password_max_age_days,
            password_max_retries=self.password_max_retries,
            password_lockout_time_mins=self.password_lockout_time_mins,
            password_history=self.password_history,
            name=self.name,
            created_on=self.created_on,
            database_name=self.database_name,
            schema_name=self.schema_name,
            owner=self.owner,
            owner_role_type=self.owner_role_type,
        )

    @classmethod
    def _from_model(cls, model) -> PasswordPolicyModel:
        return PasswordPolicyModel(
            comment=model.comment,
            password_min_length=model.password_min_length,
            password_max_length=model.password_max_length,
            password_min_upper_case_chars=model.password_min_upper_case_chars,
            password_min_lower_case_chars=model.password_min_lower_case_chars,
            password_min_numeric_chars=model.password_min_numeric_chars,
            password_min_special_chars=model.password_min_special_chars,
            password_min_age_days=model.password_min_age_days,
            password_max_age_days=model.password_max_age_days,
            password_max_retries=model.password_max_retries,
            password_lockout_time_mins=model.password_lockout_time_mins,
            password_history=model.password_history,
            name=model.name,
            created_on=model.created_on,
            database_name=model.database_name,
            schema_name=model.schema_name,
            owner=model.owner,
            owner_role_type=model.owner_role_type,
        )

    def to_dict(self):
        """Create a dictionary of the properties from a PasswordPolicy.

        This method constructs a dictionary with the key-value entries corresponding to the properties of the PasswordPolicy object.

        Returns
        _______
        dict
            A dictionary object created using the input model.
        """
        return self._to_model().to_dict()

    @classmethod
    def from_dict(cls, obj: dict) -> PasswordPolicyModel:
        """Create an instance of PasswordPolicy from a dict.

        This method constructs a PasswordPolicy object from a dictionary with the key-value pairs of its properties.

        Parameters
        ----------
        obj : dict
            A dictionary whose keys and values correspond to the properties of the resource object.

        Returns
        _______
        PasswordPolicy
            A PasswordPolicy object created using the input dictionary; this will fail if the required properties are missing.
        """
        return cls._from_model(PasswordPolicy.from_dict(obj))


PasswordPolicy._model_class = PasswordPolicyModel



import pytest

from snowflake.core.exceptions import NotFoundError
from snowflake.core.stream import Stream, StreamSourceTable
from tests.utils import random_string


@pytest.mark.min_sf_ver("8.35.0")
def test_drop(streams, src_temp_table):
    stream_name = random_string(10, "test_stream_")

    stream_handle = streams.create(
        Stream(name=stream_name, stream_source=StreamSourceTable(name=src_temp_table.name), comment="ThIs iS a ComMeNT")
    )

    stream_handle.drop()

    with pytest.raises(NotFoundError):
        stream_handle.drop()

    stream_handle.drop(if_exists=True)



from unittest import mock

import pytest

from snowflake.core import PollingOperation
from snowflake.core.cortex.search_service import QueryRequest

from ...utils import BASE_URL, extra_params


API_CLIENT_REQUEST = "snowflake.core._generated.api_client.ApiClient.request"


@pytest.fixture
def search_services(schema):
    return schema.cortex_search_services


@pytest.fixture
def search_service(search_services):
    return search_services["my_service"]


def test_search_collection(fake_root, search_services):
    args = (fake_root, "POST", BASE_URL + "/databases/my_db/schemas/my_schema/cortex-search-services/my_service:query")
    kwargs = extra_params(body={"columns": ["col1"], "limit": 10})

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = search_services.search_async("my_service", QueryRequest(columns=["col1"]))
        assert isinstance(op, PollingOperation)
        op.result()
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_search(fake_root, search_service):
    args = (fake_root, "POST", BASE_URL + "/databases/my_db/schemas/my_schema/cortex-search-services/my_service:query")
    kwargs = extra_params(body={"query": "hi", "columns": ["col1"], "limit": 10})

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = search_service.search_async("hi", ["col1"])
        assert isinstance(op, PollingOperation)
        op.result()
    mocked_request.assert_called_once_with(*args, **kwargs)



"""Manages Snowflake network rule."""

from ._generated.models import (
    NetworkRule,
)
from ._network_rule import NetworkRuleCollection, NetworkRuleResource


__all__ = [
    "NetworkRuleResource",
    "NetworkRuleCollection",
    "NetworkRule",
]



"""
Cortex Search REST API.

OpenAPI 3.0 specification for the Cortex Search REST API  # noqa: E501

The version of the OpenAPI document: 0.1.0
Contact: support@snowflake.com
Generated by: https://openapi-generator.tech

Do not edit this file manually.
"""

from __future__ import annotations

import json
import pprint
import re  # noqa: F401

from typing import Any, Optional, Union

from pydantic import BaseModel, ConfigDict, StrictFloat, StrictInt


class RankingWeights(BaseModel):
    """A model object representing the RankingWeights resource.

    Constructs an object of type RankingWeights with the provided properties.

    Parameters
    __________
    texts : float, optional
        Multiplier for text-based scoring in ranking.
    vectors : float, optional
        Multiplier for vector-based scoring in ranking.
    reranker : float, optional
        Multiplier for reranker score in ranking.
    """

    texts: Optional[Union[StrictFloat, StrictInt]] = None

    vectors: Optional[Union[StrictFloat, StrictInt]] = None

    reranker: Optional[Union[StrictFloat, StrictInt]] = None

    __properties = ["texts", "vectors", "reranker"]

    model_config = ConfigDict(
        validate_by_name=True,
        validate_assignment=True,
    )

    def to_str(self) -> str:
        """Returns the string representation of the model using alias."""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias."""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> RankingWeights:
        """Create an instance of RankingWeights from a JSON string."""
        return cls.from_dict(json.loads(json_str))

    def to_dict(
        self,
        hide_readonly_properties: bool = False,
    ) -> dict[str, Any]:
        """Returns the dictionary representation of the model using alias."""
        exclude_properties = set()

        if hide_readonly_properties:
            exclude_properties.update({})

        _dict = self.model_dump(serialize_as_any=True, by_alias=True, exclude=exclude_properties, exclude_none=True)

        return _dict

    def to_dict_without_readonly_properties(self) -> dict[str, Any]:
        """Return the dictionary representation of the model without readonly properties."""
        return self.to_dict(hide_readonly_properties=True)

    @classmethod
    def from_dict(cls, obj: dict) -> RankingWeights:
        """Create an instance of RankingWeights from a dict."""
        if obj is None:
            return None

        if type(obj) is not dict:
            return RankingWeights.model_validate(obj)

        _obj = RankingWeights.model_validate(
            {
                "texts": obj.get("texts"),
                "vectors": obj.get("vectors"),
                "reranker": obj.get("reranker"),
            }
        )

        return _obj


class RankingWeightsModel:
    def __init__(
        self,  # optional properties
        texts: Optional[float] = None,
        vectors: Optional[float] = None,
        reranker: Optional[float] = None,
    ):
        """A model object representing the RankingWeights resource.

        Constructs an object of type RankingWeights with the provided properties.

        Parameters
        __________
        texts : float, optional
            Multiplier for text-based scoring in ranking.
        vectors : float, optional
            Multiplier for vector-based scoring in ranking.
        reranker : float, optional
            Multiplier for reranker score in ranking.
        """
        self.texts = texts
        self.vectors = vectors
        self.reranker = reranker

    __properties = ["texts", "vectors", "reranker"]

    def __repr__(self) -> str:
        return repr(self._to_model())

    def _to_model(self):
        return RankingWeights(
            texts=self.texts,
            vectors=self.vectors,
            reranker=self.reranker,
        )

    @classmethod
    def _from_model(cls, model) -> RankingWeightsModel:
        return RankingWeightsModel(
            texts=model.texts,
            vectors=model.vectors,
            reranker=model.reranker,
        )

    def to_dict(self):
        """Create a dictionary of the properties from a RankingWeights.

        This method constructs a dictionary with the key-value entries corresponding to the properties of the RankingWeights object.

        Returns
        _______
        dict
            A dictionary object created using the input model.
        """
        return self._to_model().to_dict()

    @classmethod
    def from_dict(cls, obj: dict) -> RankingWeightsModel:
        """Create an instance of RankingWeights from a dict.

        This method constructs a RankingWeights object from a dictionary with the key-value pairs of its properties.

        Parameters
        ----------
        obj : dict
            A dictionary whose keys and values correspond to the properties of the resource object.

        Returns
        _______
        RankingWeights
            A RankingWeights object created using the input dictionary; this will fail if the required properties are missing.
        """
        return cls._from_model(RankingWeights.from_dict(obj))


RankingWeights._model_class = RankingWeightsModel



import contextlib
import functools
import hashlib
import logging
import os
import platform

from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Callable, Optional, TypeVar, Union

from typing_extensions import Concatenate, ParamSpec

from snowflake.connector import SnowflakeConnection
from snowflake.connector.telemetry import (
    TelemetryClient,
    TelemetryData,
)
from snowflake.connector.telemetry import (
    TelemetryField as ConnectorTelemetryField,
)
from snowflake.connector.time_util import get_time_millis

from .._common import ObjectCollection, ObjectReferenceMixin
from ..exceptions import SnowflakePythonError
from ..version import __version__ as VERSION
from .utils import TelemetryField, is_running_inside_stored_procedure


if TYPE_CHECKING:
    from ..cortex.chat_service import CortexChatService
    from ..cortex.embed_service import CortexEmbedService
    from ..cortex.inference_service import CortexInferenceService
    from ..cortex.lite_agent_service import CortexAgentService
    from ..task.dagv1 import DAGOperation


logger = logging.getLogger(__name__)

# Constant to decide whether we are running tests
_called_from_test = False

TelemetryEvent = dict[str, Any]


@dataclass
class _TelemetryEventBuilder:
    class_name: str
    func_name: str
    _type: Optional[str] = None
    _data: Optional[dict[str, Any]] = None

    @staticmethod
    def _get_ci_environment_type() -> str:
        if "SF_GITHUB_ACTION" in os.environ:
            return "SF_GITHUB_ACTION"
        if "GITHUB_ACTIONS" in os.environ:
            return "GITHUB_ACTIONS"
        if "GITLAB_CI" in os.environ:
            return "GITLAB_CI"
        if "CIRCLECI" in os.environ:
            return "CIRCLECI"
        if "JENKINS_URL" in os.environ or "HUDSON_URL" in os.environ:
            return "JENKINS"
        if "TF_BUILD" in os.environ:
            return "AZURE_DEVOPS"
        return "UNKNOWN"

    def _build(self) -> dict[str, Any]:
        if self._type is None:
            raise ValueError("event type not set")
        return {
            ConnectorTelemetryField.KEY_SOURCE.value: "snowflake.core",
            TelemetryField.KEY_VERSION.value: VERSION,
            TelemetryField.KEY_PYTHON_VERSION.value: platform.python_version(),
            TelemetryField.KEY_OS.value: platform.system(),
            ConnectorTelemetryField.KEY_TYPE.value: self._type,
            TelemetryField.KEY_CI_ENVIRONMENT_TYPE.value: self._get_ci_environment_type(),
            TelemetryField.KEY_DATA.value: {
                "class_name": self.class_name,
                TelemetryField.KEY_FUNC_NAME.value: self.func_name,
                **(self._data if self._data else {}),
            },
        }

    def usage_event(self) -> TelemetryEvent:
        self._type = "python_api"
        return self._build()

    def exception_event(self, exception: Exception) -> TelemetryEvent:
        from snowflake.core.exceptions import APIError

        self._type = "python_api_exception"
        self._data = {
            "exception_type": type(exception).__name__,
            "exception_sha256": hashlib.sha256(str(exception).encode()).hexdigest(),
            "is_python_api_error": isinstance(exception, SnowflakePythonError),
        }
        if isinstance(exception, APIError):
            request_info = exception.get_request_info()
            self._data["http_code"] = exception.status
            self._data["request_id"] = request_info["request_id"]
            self._data["error_code"] = request_info["error_code"]
        return self._build()


class ApiTelemetryClient:
    def __init__(self, conn: SnowflakeConnection) -> None:
        self.telemetry: Optional[TelemetryClient] = None if is_running_inside_stored_procedure() else conn._telemetry
        logger.info("telemetry client created for %r, telemetry enabled: %s", conn, bool(self.telemetry))

    def _send(self, msg: TelemetryEvent, timestamp: Optional[int] = None) -> None:
        if not self.telemetry:
            return
        if not timestamp:
            timestamp = get_time_millis()
        telemetry_data = TelemetryData(message=msg, timestamp=timestamp)
        self.telemetry.try_add_log_to_batch(telemetry_data)

    def safe_send(
        self,
        data: TelemetryEvent,
    ) -> None:
        with contextlib.suppress(Exception):
            if not self.telemetry:
                return
            self._send(data)


P = ParamSpec("P")
R = TypeVar("R")


def api_telemetry(func: Callable[Concatenate[Any, P], R]) -> Callable[Concatenate[Any, P], R]:
    @functools.wraps(func)
    def wrap(
        self: Union[
            ObjectReferenceMixin[Any],
            ObjectCollection[Any],
            "DAGOperation",
            "CortexInferenceService",
            "CortexChatService",
            "CortexEmbedService",
            "CortexAgentService",
        ],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> R:
        from ..cortex.chat_service import CortexChatService
        from ..cortex.embed_service import CortexEmbedService
        from ..cortex.inference_service import CortexInferenceService
        from ..cortex.lite_agent_service import CortexAgentService
        from ..task.dagv1 import DAGOperation

        if isinstance(self, (ObjectReferenceMixin, ObjectCollection)):
            telemetry_client = self.root._telemetry_client  # type: ignore
        elif isinstance(self, DAGOperation):
            telemetry_client = self.schema.root._telemetry_client
        elif isinstance(self, (CortexChatService, CortexInferenceService, CortexEmbedService, CortexAgentService)):
            telemetry_client = self._api._root._telemetry_client
        else:
            raise TypeError(f"unknown type {type(self)}")
        api = None
        if hasattr(self, "_api"):
            api = self._api
        elif hasattr(self, "collection") and hasattr(self.collection, "_api"):
            api = self.collection._api
        elif _called_from_test and not isinstance(self, DAGOperation):
            # DAGOperation will not be reported when the API object cannot be extracted
            #  from them. This is okay because this class will call other APIs
            #  downstream.
            raise Exception(f"cannot determine API for {self=}")
        if api is not None:
            # Cause resolution of api client, if not done beforehand
            api.api_client  # noqa: B018
        class_name = self.__class__.__name__
        func_name = func.__name__
        logger.debug(
            "calling method %s on class %s after submitting telemetry if enabled",
            func_name,
            class_name,
        )

        event_builder = _TelemetryEventBuilder(class_name=class_name, func_name=func_name)

        telemetry_client.safe_send(event_builder.usage_event())
        try:
            r = func(self, *args, **kwargs)
            return r
        except Exception as err:
            try:
                telemetry_client.safe_send(event_builder.exception_event(err))
            except Exception as telemetry_err:
                logging.debug("Failed to send telemetry: %s", telemetry_err)
            raise

    return wrap



import pytest

from snowflake.core.exceptions import NotFoundError
from snowflake.core.streamlit import Streamlit
from tests.integ.utils import random_string


@pytest.mark.min_sf_ver("9.38.0")
def test_drop(streamlits, streamlit_stage_with_file, warehouse, streamlit_main_file):
    name = random_string(8, "test_streamlit_")
    st = Streamlit(
        name=name,
        query_warehouse=warehouse.name,
        source_location=f"@{streamlit_stage_with_file.name}",
        main_file=streamlit_main_file,
    )

    ref = streamlits.create(st)
    ref.drop()

    with pytest.raises(NotFoundError):
        streamlits[name].fetch()

    # drop if exists shouldn't error
    streamlits[name].drop(if_exists=True)



import copy

import pytest

from snowflake.core.exceptions import APIError, ConflictError, NotFoundError
from tests.integ.utils import random_string

from .conftest import test_tag_template


def test_drop(tags):
    prefix = "test_tag_drop_"
    pre_create_count = len(list(tags.iter(like=prefix + "%")))

    name = random_string(10, prefix)
    tag_handle = tags[name]

    tag = copy.deepcopy(test_tag_template)
    tag.name = name
    tags.create(tag)

    created_count = len(list(tags.iter(like=prefix + "%")))
    assert pre_create_count + 1 == created_count

    tag_handle.drop()

    after_drop_count = len(list(tags.iter(like=prefix + "%")))
    assert pre_create_count == after_drop_count

    with pytest.raises(NotFoundError):
        tag_handle.drop()
        tag_handle.drop(if_exists=False)

    tag_handle.drop(if_exists=True)


def test_undrop(tags):
    name = random_string(10, "test_tag_drop_undrop_")
    tag_handle = tags[name]

    tag = copy.deepcopy(test_tag_template)
    tag.name = name

    # attempting to undrop when the object never existed results in 400
    with pytest.raises(APIError):
        tag_handle.undrop()

    tags.create(tag)

    tag_handle.drop()
    tag_handle.undrop()
    assert tag_handle.fetch() is not None

    # attempting to undrop an existing object results in 409
    with pytest.raises(ConflictError):
        tag_handle.undrop()

    tag_handle.drop(if_exists=True)



"""
Snowflake External Volume API.

The Snowflake External Volume API is a REST API that you can use to access, update, and perform certain actions on External Volume resource in a Snowflake database.  # noqa: E501

The version of the OpenAPI document: 0.0.1
Contact: support@snowflake.com
Generated by: https://openapi-generator.tech

Do not edit this file manually.
"""

from __future__ import annotations

import json
import pprint
import re  # noqa: F401

from typing import Any, Optional

from pydantic import ConfigDict, StrictStr

from snowflake.core.external_volume._generated.models.encryption import Encryption, EncryptionModel
from snowflake.core.external_volume._generated.models.storage_location import StorageLocation


class StorageLocationGcs(StorageLocation):
    """A model object representing the StorageLocationGcs resource.

    Constructs an object of type StorageLocationGcs with the provided properties.

    Parameters
    __________

    name : str

    storage_base_url : str
        Specifies the base URL for your cloud storage location.
    encryption : Encryption, optional
    """

    storage_base_url: StrictStr

    encryption: Optional[Encryption] = None

    __properties = ["name", "storage_provider"]

    model_config = ConfigDict(
        validate_by_name=True,
        validate_assignment=True,
    )

    def to_str(self) -> str:
        """Returns the string representation of the model using alias."""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias."""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> StorageLocationGcs:
        """Create an instance of StorageLocationGcs from a JSON string."""
        return cls.from_dict(json.loads(json_str))

    def to_dict(
        self,
        hide_readonly_properties: bool = False,
    ) -> dict[str, Any]:
        """Returns the dictionary representation of the model using alias."""
        exclude_properties = set()

        if hide_readonly_properties:
            exclude_properties.update({})

        _dict = self.model_dump(serialize_as_any=True, by_alias=True, exclude=exclude_properties, exclude_none=True)

        # override the default output from pydantic by calling `to_dict()` of encryption
        if self.encryption:
            _dict["encryption"] = self.encryption.to_dict()

        _dict["storage_provider"] = StorageLocation.get_child_model_discriminator_value("StorageLocationGcs")

        return _dict

    def to_dict_without_readonly_properties(self) -> dict[str, Any]:
        """Return the dictionary representation of the model without readonly properties."""
        return self.to_dict(hide_readonly_properties=True)

    @classmethod
    def from_dict(cls, obj: dict) -> StorageLocationGcs:
        """Create an instance of StorageLocationGcs from a dict."""
        if obj is None:
            return None

        if type(obj) is not dict:
            return StorageLocationGcs.model_validate(obj)

        _obj = StorageLocationGcs.model_validate(
            {
                "name": obj.get("name"),
                "storage_base_url": obj.get("storage_base_url"),
                "encryption": Encryption.from_dict(obj.get("encryption"))
                if obj.get("encryption") is not None
                else None,
            }
        )

        return _obj


class StorageLocationGcsModel(StorageLocation):
    def __init__(
        self,
        name: str,
        storage_base_url: str,
        # optional properties
        encryption: Optional[Encryption] = None,
    ):
        """A model object representing the StorageLocationGcs resource.

        Constructs an object of type StorageLocationGcs with the provided properties.

        Parameters
        __________

        name : str

        storage_base_url : str
            Specifies the base URL for your cloud storage location.
        encryption : Encryption, optional
        """
        super().__init__(
            name=name,
        )
        self.storage_base_url = storage_base_url
        self.encryption = encryption

    __properties = ["name", "storage_provider"]

    def __repr__(self) -> str:
        return repr(self._to_model())

    def _to_model(self):
        return StorageLocationGcs(
            name=self.name,
            storage_base_url=self.storage_base_url,
            encryption=self.encryption._to_model() if self.encryption is not None else None,
        )

    @classmethod
    def _from_model(cls, model) -> StorageLocationGcsModel:
        return StorageLocationGcsModel(
            name=model.name,
            storage_base_url=model.storage_base_url,
            encryption=EncryptionModel._from_model(model.encryption) if model.encryption else None,
        )

    def to_dict(self):
        """Create a dictionary of the properties from a StorageLocationGcs.

        This method constructs a dictionary with the key-value entries corresponding to the properties of the StorageLocationGcs object.

        Returns
        _______
        dict
            A dictionary object created using the input model.
        """
        return self._to_model().to_dict()

    @classmethod
    def from_dict(cls, obj: dict) -> StorageLocationGcsModel:
        """Create an instance of StorageLocationGcs from a dict.

        This method constructs a StorageLocationGcs object from a dictionary with the key-value pairs of its properties.

        Parameters
        ----------
        obj : dict
            A dictionary whose keys and values correspond to the properties of the resource object.

        Returns
        _______
        StorageLocationGcs
            A StorageLocationGcs object created using the input dictionary; this will fail if the required properties are missing.
        """
        return cls._from_model(StorageLocationGcs.from_dict(obj))


StorageLocationGcs._model_class = StorageLocationGcsModel
